#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/10/07 09:06

import re

class RgbFormatter:
    """ Format Color """
    def __init__(self, r=0, g=0, b=0):
        """ 标准的RGB颜色表达方式，数值范围在0~255 """
        self.r, self.g, self.b = r, g, b

    def __repr__(self):
        return f'({self.r}, {self.g}, {self.b})'

    def mixtures(self, rgbs, ratios=1):
        """ 混入另一部分颜色，产生一个新的颜色

        :param rgbs: 新的颜色
        :param ratios: 原颜色权重是1，其他颜色默认权重也是1

        >>> RgbFormatter(255, 0, 0).mixtures([RgbFormatter(0, 255, 0), RgbFormatter(0, 0, 255)])
        (85, 85, 85)
        """
        # 1 非列表结构全部转列表
        if isinstance(rgbs, RgbFormatter):
            rgbs = [rgbs]
        if not isinstance(ratios, (list, tuple)):
            ratios = [ratios]

        # 2 添加当前自身颜色
        rgbs.append(self)
        ratios.append(1)
        # ratios数与rgbs数相同
        if len(ratios) < len(rgbs):
            ratios += [1] * (len(rgbs) - len(ratios))

        # 3 权重计算
        sum_ratio = sum(ratios)
        r = sum([rgb.r * ratio for rgb, ratio in zip(rgbs, ratios)]) / sum_ratio
        g = sum([rgb.g * ratio for rgb, ratio in zip(rgbs, ratios)]) / sum_ratio
        b = sum([rgb.b * ratio for rgb, ratio in zip(rgbs, ratios)]) / sum_ratio
        return RgbFormatter(round(r), round(g), round(b))

    def light(self, ratio=1):
        """ 把颜色变浅

        :param ratio: 原值权重是1，这里可以设置白色的权重
        :return:

        >>> RgbFormatter(0, 0, 0).light()
        (128, 128, 128)
        """
        return self.mixtures(RgbFormatter(255, 255, 255), ratio)

    @staticmethod
    def from_rgb_int255(r=0, g=0, b=0):
        return RgbFormatter(r, g, b)

    def to_tuple(self):
        return self.r, self.g, self.b

    @staticmethod
    def from_vba_value(v):
        """ VBA使用一个Long类型存储颜色，像素(b,g,r) = b*65536 + g*256 + r

        >>> RgbFormatter().from_vba_value(11278293)
        (213, 23, 172)

        # 有时候用宏生成代码，可能会产生负值
        >>> RgbFormatter().from_vba_value(-11260549)
        (123, 45, 84)
        >>> RgbFormatter().from_vba_value(-16776961)
        (255, 0, 0)
        """
        if v < 0: v += 256 ** 3
        rgb = [v % 256, (v // 256) % 256, (v // 65536)]
        return RgbFormatter(*rgb)

    def to_vba_value(self, negative=False):
        """ 转vba的颜色代号值

        :param negative: 默认返回正数，可以设置获得负数表达的数值

        >>> RgbFormatter(123, 45, 84).to_vba_value()
        5516667
        >>> RgbFormatter(123, 45, 84).to_vba_value(negative=True)
        -11260549
        """
        v = self.r + self.g * 256 + self.b * 65536
        if negative: v -= 256 ** 3
        return v

    @property
    def vba_value(self):
        return self.to_vba_value()

    @staticmethod
    def from_hex(s):
        """
        >>> RgbFormatter.from_hex('#7B2D54')
        (123, 45, 84)
        >>> RgbFormatter.from_hex('7B2D54')
        (123, 45, 84)
        """
        m = re.search(r'(\w\w)(\w\w)(\w\w)', s)
        if not m: raise ValueError(f'16进制颜色格式有误: {s}')
        return RgbFormatter(*[int(x, 16) for x in m.groups()])

    def to_hex(self, lower=False):
        """
        :param lower: 默认返回大写，可以设置返回小写

        >>> RgbFormatter(123, 45, 84).to_hex()
        '#7B2D54'
        """
        pattern = '#{:02X}{:02X}{:02X}'
        if lower: pattern = pattern.replace('X', 'x')
        return pattern.format(self.r, self.g, self.b)

    @property
    def hex(self):
        return self.to_hex()

    @staticmethod
    def from_percentage(r, g, b):
        """ 主要用于svg等场景会用的百分率的格式
        输入的每个值是0~1内的百分率浮点数
        """
        return RgbFormatter(*[round(255 * x) for x in (r, g, b)])

    def to_percentage(self):
        """
        >>> RgbFormatter.from_percentage(0.5, 0.7, 0.9).to_percentage()
        [0.5019607843137255, 0.6980392156862745, 0.9019607843137255]
        """
        return [x / 255 for x in (self.r, self.g, self.b)]

    @property
    def percentage(self):
        return self.to_percentage()



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

import os
import sys
import json

__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '..')))

os.environ["FLAGS_allocator_strategy"] = 'auto_growth'

import paddle

from ppocr.data import create_operators, transform
from ppocr.modeling.architectures import build_model
from ppocr.postprocess import build_post_process
from ppocr.utils.save_load import load_model
from ppocr.utils.utility import get_image_file_list
import tools.program as program


def main():
    global_config = config['Global']

    # build post process
    post_process_class = build_post_process(config['PostProcess'],
                                            global_config)

    # build model
    if hasattr(post_process_class, 'character'):
        char_num = len(getattr(post_process_class, 'character'))
        if config['Architecture']["algorithm"] in ["Distillation",
                                                   ]:  # distillation model
            for key in config['Architecture']["Models"]:
                config['Architecture']["Models"][key]["Head"][
                    'out_channels'] = char_num
        else:  # base rec model
            config['Architecture']["Head"]['out_channels'] = char_num

    model = build_model(config['Architecture'])

    load_model(config, model)

    # create data ops
    transforms = []
    for op in config['Eval']['dataset']['transforms']:
        op_name = list(op)[0]
        if 'Label' in op_name:
            continue
        elif op_name in ['RecResizeImg']:
            op[op_name]['infer_mode'] = True
        elif op_name == 'KeepKeys':
            if config['Architecture']['algorithm'] == "SRN":
                op[op_name]['keep_keys'] = [
                    'image', 'encoder_word_pos', 'gsrm_word_pos',
                    'gsrm_slf_attn_bias1', 'gsrm_slf_attn_bias2'
                ]
            elif config['Architecture']['algorithm'] == "SAR":
                op[op_name]['keep_keys'] = ['image', 'valid_ratio']
            else:
                op[op_name]['keep_keys'] = ['image']
        transforms.append(op)
    global_config['infer_mode'] = True
    ops = create_operators(transforms, global_config)

    save_res_path = config['Global'].get('save_res_path',
                                         "./output/rec/predicts_rec.txt")
    if not os.path.exists(os.path.dirname(save_res_path)):
        os.makedirs(os.path.dirname(save_res_path))

    model.eval()

    with open(save_res_path, "w") as fout:
        for file in get_image_file_list(config['Global']['infer_img']):
            logger.info("infer_img: {}".format(file))
            with open(file, 'rb') as f:
                img = f.read()
                data = {'image': img}
            batch = transform(data, ops)
            if config['Architecture']['algorithm'] == "SRN":
                encoder_word_pos_list = np.expand_dims(batch[1], axis=0)
                gsrm_word_pos_list = np.expand_dims(batch[2], axis=0)
                gsrm_slf_attn_bias1_list = np.expand_dims(batch[3], axis=0)
                gsrm_slf_attn_bias2_list = np.expand_dims(batch[4], axis=0)

                others = [
                    paddle.to_tensor(encoder_word_pos_list),
                    paddle.to_tensor(gsrm_word_pos_list),
                    paddle.to_tensor(gsrm_slf_attn_bias1_list),
                    paddle.to_tensor(gsrm_slf_attn_bias2_list)
                ]
            if config['Architecture']['algorithm'] == "SAR":
                valid_ratio = np.expand_dims(batch[-1], axis=0)
                img_metas = [paddle.to_tensor(valid_ratio)]

            images = np.expand_dims(batch[0], axis=0)
            images = paddle.to_tensor(images)
            if config['Architecture']['algorithm'] == "SRN":
                preds = model(images, others)
            elif config['Architecture']['algorithm'] == "SAR":
                preds = model(images, img_metas)
            else:
                preds = model(images)
            post_result = post_process_class(preds)
            info = None
            if isinstance(post_result, dict):
                rec_info = dict()
                for key in post_result:
                    if len(post_result[key][0]) >= 2:
                        rec_info[key] = {
                            "label": post_result[key][0][0],
                            "score": float(post_result[key][0][1]),
                        }
                info = json.dumps(rec_info)
            else:
                if len(post_result[0]) >= 2:
                    info = post_result[0][0] + "\t" + str(post_result[0][1])

            if info is not None:
                logger.info("\t result: {}".format(info))
                fout.write(file + "\t" + info)
    logger.info("success!")


if __name__ == '__main__':
    config, device, logger, vdl_writer = program.preprocess()
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/05/26 17:24

from pyxllib.prog.pupil import check_install_package

check_install_package('qtpy', 'QtPy')

import json
import os.path as osp
import sys
import time

from PyQt5.QtCore import pyqtSignal
from qtpy import QtWidgets
from qtpy import QtGui
from qtpy.QtWidgets import QFrame, QInputDialog, QApplication

from pyxllib.prog.newbie import CvtType

here = osp.dirname(osp.abspath(__file__))


class QHLine(QFrame):
    """ https://stackoverflow.com/questions/5671354/how-to-programmatically-make-a-horizontal-line-in-qt """

    def __init__(self):
        super(QHLine, self).__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)


class QVLine(QFrame):
    def __init__(self):
        super(QVLine, self).__init__()
        self.setFrameShape(QFrame.VLine)
        self.setFrameShadow(QFrame.Sunken)


class XlLineEdit(QtWidgets.QLineEdit):
    correctChanged = pyqtSignal(object)  # 符合数值类型的修改
    wrongChanged = pyqtSignal(str)  # 不符合数值类型的修改

    def __init__(self, text=None, parent=None, *, valcvt=None):
        """
        :param valcvt: 数值类型转换器
        """
        super().__init__(str(text), parent)

        def check():
            # TODO 目前是强制重置样式，可以考虑怎么保留原样式基础上修改属性值
            s = self.text()
            try:
                if valcvt:
                    s = valcvt(s)
                self.setStyleSheet('')
                self.setToolTip('')
                self.correctChanged.emit(s)
            except ValueError:
                self.setStyleSheet('background-color: lightpink;')
                self.setToolTip(f'输入数据不是{valcvt}类型')
                self.wrongChanged.emit(s)

        self.textChanged.connect(check)
        if text:
            check()

        # self.setStyleSheet(self.styleSheet() + 'qproperty-cursorPosition: 0;')
        self.setStyleSheet(self.styleSheet())


class XlComboBox(QtWidgets.QComboBox):
    # 这个控件一般没有类型检查，但在支持填入自定义值时，是存在类型错误问题的
    #   但在工程上还是依然写了 correctChanged，方便下游任务统一接口
    correctChanged = pyqtSignal(object)  # 符合数值类型的修改
    wrongChanged = pyqtSignal(str)  # 不符合数值类型的修改

    def __init__(self, parent=None, *, text=None, items=None, valcvt=None, editable=False):
        """
        """
        # 1 基础设置
        super().__init__(parent)
        self.reset_items(items)
        self.editable = editable
        if self.editable:
            self.setEditable(True)

        # 2 检查功能
        def check(s):
            try:
                if valcvt:
                    s = valcvt(s)

                if self.editable:  # 支持自定义值
                    self.setStyleSheet('')
                    self.setToolTip('')
                    self.correctChanged.emit(s)
                elif s not in self.items_set:  # 不支持自定义值，但出现自定义值
                    self.setStyleSheet('background-color: yellow;')
                    self.setToolTip(f'不在清单里的非法值')
                    self.wrongChanged.emit(s)
                else:  # 不支持自定义值，且目前值在清单中
                    self.setEditable(False)
                    self.setStyleSheet('')
                    self.setToolTip('')
                    self.correctChanged.emit(s)
            except ValueError:
                self.setStyleSheet('background-color: lightpink;')
                self.setToolTip(f'输入数据不是{valcvt}类型')
                self.wrongChanged.emit(s)

        self.currentTextChanged.connect(check)
        # self.wrongChanged.connect(lambda s: print('非法值：', s))  # 可以监控非法值

        # 3 是否有预设值
        if text:
            self.setText(text)

        # 4 补充格式
        # self.setStyleSheet(self.styleSheet() + 'qproperty-cursorPosition: 0;')
        self.setStyleSheet(self.styleSheet())

    def setText(self, text):
        text = str(text)
        if text not in self.items and not self.editable:
            # 虽然不支持editable，但是出现了意外值，需要强制升级为可编辑
            self.setEditable(True)
        self.setCurrentText(text)

    def reset_items(self, items):
        # 1 存储配置清单
        self.clear()
        self.raw_items = items  # noqa
        self.items = [str(x) for x in items if x is not None]  # noqa
        self.items_set = set(self.items)  # noqa 便于判断是否存在的集合类型
        self.addItems(self.items)

        # 2 画出 元素值、分隔符
        cnt = 0
        for i in range(len(items)):
            if items[i] is None:
                self.insertSeparator(i - cnt)  # 不过这个分割符没那么显眼
                cnt += 1


def get_input_widget(items=None, cur_value=None, *, valcvt=None,
                     n_widget=1, enabled=True,
                     correct_changed=None):
    """ 根据items参数情况，智能判断生成对应的widget

    :param items:
        None, 普通的文本编辑框
        普通数组，下拉框  （可以用None元素表示分隔符）
            list，除了列表中枚举值，也支持自定义输入其他值
            tuple，只能用列表中的枚举值
        多级嵌套数组，多级下拉框  （未实装）  （一般都是不可改的，并且同类型的数据）
            [('福建', [('龙岩', ['连城', '长汀', ...], ...)]), ('北京', ...)]
            这种情况会返回多个widget
    :param cur_value: 当前显示的文本值
    :param valcvt: 数值类型转换函数，非法时返回ValueError  （未实装）
        很多输入框是传入文本，有时需要转为int、float、list等类型
        支持输入常见类型转换的字符串名称，比如int、float
    :param correct_changed: 文本改变时的回调函数
    :param n_widget: 配合items为嵌套数组使用，需要指定嵌套层数
        此时cur_value、cvt、enabled、text_changed等系列值可以传入n_widget长度的list
    :param enabled: 是否可编辑
    """
    if n_widget > 1:
        raise NotImplementedError

    # 1 封装类型检查功能
    if isinstance(valcvt, str):
        cvtfunc = CvtType.factory(valcvt)
    else:
        cvtfunc = valcvt

    # 2 正式生成控件
    if isinstance(items, (list, tuple)):
        # 带有 items 的字段支持候选下拉菜单
        w = XlComboBox(text=cur_value, items=items, valcvt=cvtfunc, editable=isinstance(items, list))
    elif items is None:
        # 普通填充框
        w = XlLineEdit(cur_value, valcvt=cvtfunc)
    else:
        raise ValueError(f'{type(items)}')

    # 3 通用配置
    if callable(correct_changed):
        w.correctChanged.connect(correct_changed)
    if not enabled:
        w.setEnabled(enabled)

    return w


def newIcon(icon):
    icons_dir = osp.join(here, "../icons")
    return QtGui.QIcon(osp.join(":/", icons_dir, "%s.png" % icon))


def newAction(
        parent,
        text,
        slot=None,
        shortcut=None,
        icon=None,
        tip=None,
        checkable=False,
        enabled=True,
        checked=False,
):
    """Create a new action and assign callbacks, shortcuts, etc."""
    a = QtWidgets.QAction(text, parent)
    if icon is not None:
        a.setIconText(text.replace(" ", "\n"))
        a.setIcon(newIcon(icon))
    if shortcut is not None:
        if isinstance(shortcut, (list, tuple)):
            a.setShortcuts(shortcut)
        else:
            a.setShortcut(shortcut)
    if tip is not None:
        a.setToolTip(tip)
        a.setStatusTip(tip)
    if slot is not None:
        a.triggered.connect(slot)
    if checkable:
        a.setCheckable(True)
    a.setEnabled(enabled)
    a.setChecked(checked)
    return a


class XlActionFunc:
    """ 跟action有关的函数调用封装 （该类也可以作为基础的check版本使用）

    一般逻辑结构是这样：
        有个可运行功能的函数func，运行时会配置一些需要存储起来的变量值value
        并且功能需要关联action，绑定到menu等菜单中时，可以使用该装饰器
    """

    def __init__(self, parent, title, value=None, checked=None, **kwargs):
        self.parent = parent
        self.title = title
        self.value = value
        self.checked = bool(checked)
        self.checkable = checked is not None

        self.action = newAction(self.parent, self.parent.tr(self.title), self.__call__,
                                checkable=self.checkable, checked=self.checked, **kwargs)

    def __call__(self, checked):
        self.checked = checked


class GetMultiLineTextAction(XlActionFunc):
    """ 该类value是直接存储原始的完整文本内容 """

    def __call__(self, checked):
        super().__call__(checked)
        self.value = self.value or ''
        inputs = QInputDialog.getMultiLineText(self.parent, self.title, '编辑文本：', self.value)
        if inputs[1]:  # “确定操作” 才更新属性
            self.value = inputs[0]


class GetItemsAction(XlActionFunc):
    """ 该类value目前是存储为list类型 """

    def __call__(self, checked):
        super().__call__(checked)
        self.value = self.value or []
        inputs = QInputDialog.getMultiLineText(self.parent, self.title, '编辑多行文本：',
                                               '\n'.join(self.value))
        if inputs[1]:  # “确定操作” 才更新属性
            self.value = inputs[0].splitlines()


class GetJsonAction(XlActionFunc):
    """ 该类value是直接存储原始的完整文本内容 """

    def __call__(self, checked):
        super().__call__(checked)
        self.value = self.value or ''
        inputs = QInputDialog.getMultiLineText(self.parent, self.title, '编辑json数据：',
                                               json.dumps(self.value, indent=2))
        if inputs[1]:  # “确定操作” 才更新属性
            self.value = json.loads(inputs[0])


def main_qapp(window):
    """ 执行Qt应用 """
    app = QApplication(sys.argv)
    window.show()  # 展示窗口
    sys.exit(app.exec_())


def qt_clipboard_monitor(func=None, verbose=1, *, cooldown=0.5):
    """ qt实现的剪切板监控器

    :param cooldown: cd，冷切时间，防止短时间内因为重复操作响应剪切板，重复执行功能

    感觉这个组件还有很多可以扩展的，比如设置可以退出的快捷键
    """
    import pyperclip

    last_response = time.time()

    if func is None:
        func = lambda s: s

    def on_clipboard_change():
        # 1 数据内容一样则跳过不处理，表示很可能是该函数调用pyperclip.copy(s)产生的重复响应
        nonlocal last_response
        s0 = pyperclip.paste()
        s0 = s0.replace('\r\n', '\n')

        cur_time = time.time()

        if cur_time - last_response < cooldown:
            return
        last_response = cur_time

        # 2 处理函数
        s1 = func(s0)
        if s1 != s0:
            if verbose:
                print('【处理前】', time.strftime('%H:%M:%S'))
                print(s0)
                print('【处理后】')
                print(s1)
                print()
            pyperclip.copy(s1)

    app = QApplication([])
    clipboard = app.clipboard()
    clipboard.dataChanged.connect(on_clipboard_change)
    app.exec_()



# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
This code is refer from:
https://github.com/whai362/PSENet/blob/python3/models/head/psenet_head.py
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import cv2
import paddle
from paddle.nn import functional as F

from ppocr.postprocess.pse_postprocess.pse import pse


class PSEPostProcess(object):
    """
    The post process for PSE.
    """

    def __init__(self,
                 thresh=0.5,
                 box_thresh=0.85,
                 min_area=16,
                 box_type='box',
                 scale=4,
                 **kwargs):
        assert box_type in ['box', 'poly'], 'Only box and poly is supported'
        self.thresh = thresh
        self.box_thresh = box_thresh
        self.min_area = min_area
        self.box_type = box_type
        self.scale = scale

    def __call__(self, outs_dict, shape_list):
        pred = outs_dict['maps']
        if not isinstance(pred, paddle.Tensor):
            pred = paddle.to_tensor(pred)
        pred = F.interpolate(
            pred, scale_factor=4 // self.scale, mode='bilinear')

        score = F.sigmoid(pred[:, 0, :, :])

        kernels = (pred > self.thresh).astype('float32')
        text_mask = kernels[:, 0, :, :]
        kernels[:, 0:, :, :] = kernels[:, 0:, :, :] * text_mask

        score = score.numpy()
        kernels = kernels.numpy().astype(np.uint8)

        boxes_batch = []
        for batch_index in range(pred.shape[0]):
            boxes, scores = self.boxes_from_bitmap(score[batch_index],
                                                   kernels[batch_index],
                                                   shape_list[batch_index])

            boxes_batch.append({'points': boxes, 'scores': scores})
        return boxes_batch

    def boxes_from_bitmap(self, score, kernels, shape):
        label = pse(kernels, self.min_area)
        return self.generate_box(score, label, shape)

    def generate_box(self, score, label, shape):
        src_h, src_w, ratio_h, ratio_w = shape
        label_num = np.max(label) + 1

        boxes = []
        scores = []
        for i in range(1, label_num):
            ind = label == i
            points = np.array(np.where(ind)).transpose((1, 0))[:, ::-1]

            if points.shape[0] < self.min_area:
                label[ind] = 0
                continue

            score_i = np.mean(score[ind])
            if score_i < self.box_thresh:
                label[ind] = 0
                continue

            if self.box_type == 'box':
                rect = cv2.minAreaRect(points)
                bbox = cv2.boxPoints(rect)
            elif self.box_type == 'poly':
                box_height = np.max(points[:, 1]) + 10
                box_width = np.max(points[:, 0]) + 10

                mask = np.zeros((box_height, box_width), np.uint8)
                mask[points[:, 1], points[:, 0]] = 255

                contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL,
                                               cv2.CHAIN_APPROX_SIMPLE)
                bbox = np.squeeze(contours[0], 1)
            else:
                raise NotImplementedError

            bbox[:, 0] = np.clip(np.round(bbox[:, 0] / ratio_w), 0, src_w)
            bbox[:, 1] = np.clip(np.round(bbox[:, 1] / ratio_h), 0, src_h)
            boxes.append(bbox)
            scores.append(score_i)
        return boxes, scores



# Copyright (c) 2021 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import paddle

__all__ = ['KIEMetric']


class KIEMetric(object):
    def __init__(self, main_indicator='hmean', **kwargs):
        self.main_indicator = main_indicator
        self.reset()
        self.node = []
        self.gt = []

    def __call__(self, preds, batch, **kwargs):
        nodes, _ = preds
        gts, tag = batch[4].squeeze(0), batch[5].tolist()[0]
        gts = gts[:tag[0], :1].reshape([-1])
        self.node.append(nodes.numpy())
        self.gt.append(gts)
        # result = self.compute_f1_score(nodes, gts)
        # self.results.append(result)

    def compute_f1_score(self, preds, gts):
        ignores = [0, 2, 4, 6, 8, 10, 12, 14, 16, 18, 20, 22, 24, 25]
        C = preds.shape[1]
        classes = np.array(sorted(set(range(C)) - set(ignores)))
        hist = np.bincount(
            (gts * C).astype('int64') + preds.argmax(1), minlength=C
            **2).reshape([C, C]).astype('float32')
        diag = np.diag(hist)
        recalls = diag / hist.sum(1).clip(min=1)
        precisions = diag / hist.sum(0).clip(min=1)
        f1 = 2 * recalls * precisions / (recalls + precisions).clip(min=1e-8)
        return f1[classes]

    def combine_results(self, results):
        node = np.concatenate(self.node, 0)
        gts = np.concatenate(self.gt, 0)
        results = self.compute_f1_score(node, gts)
        data = {'hmean': results.mean()}
        return data

    def get_metric(self):

        metircs = self.combine_results(self.results)
        self.reset()
        return metircs

    def reset(self):
        self.results = []  # clear results
        self.node = []
        self.gt = []



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/08/31 09:56

from pyxllib.prog.pupil import check_install_package

check_install_package('win32com', 'pypiwin32')

import json
import os
import re

import pythoncom
from win32com.client import constants
import win32com.client as win32

from pyxllib.prog.pupil import DictTool, EnchantBase, EnchantCvt, run_once
from pyxllib.text.pupil import strwidth
from pyxllib.debug.specialist import File, Dir, get_etag, browser


def __docx():
    """ python-docx 相关封装
    """
    pass


class DocxTools:
    @classmethod
    def to_pdf(cls, docx_file, pdf_file=None):
        check_install_package('docx2pdf')  # 安装不成功的时候可以考虑加参数：--user
        import docx2pdf

        if pdf_file is None:
            pdf_file = docx_file.with_suffix('.pdf')

        docx2pdf.convert(str(docx_file), str(pdf_file))
        return pdf_file

    @classmethod
    def merge(cls, master_file, toc, *, outline='demote'):
        """ 合并多份docx文件

        :param master_file: 要合并到哪个主文件
            注意如果这个文件已存在，会被替换，重置
        :param toc: 类似fitz的table of contents，用一个n*3的表格表示新文档的格式
            list，每个元素三列：标题级别，标题名称，(可选)对应文件内容
        :param outline: 原来每份子文档里的标题，插入到新文档中的处理规则
            demote：降级
            clear：清除

        这个功能还有些局限性，后面要扩展鲁棒性
        TODO 增加一个支持将原文档标题降级的功能，降到toc之后
        """
        app = EnchantWin32WordApplication.get_app()

        master_doc = app.new_doc(master_file)
        for item in toc:
            lvl, title, file = item

            # 1 加一个标题
            r = master_doc.Paragraphs.Add().Range
            r.InsertBefore(title)
            r.Style = master_doc.Styles(f'标题 {lvl}')

            # 2 拷贝完整的内容
            if file:
                file = File(file)
                member_doc = app.open_doc(file)
                member_doc.Activate()

                # 处理原来文档的目录级别
                if outline == 'demote':
                    member_doc.outline_demote(lvl)
                elif outline == 'clear':
                    # 降10级就相当于清除所有原来的标题
                    member_doc.outline_demote(10)

                app.Selection.WholeStory()
                app.Selection.Copy()
                master_doc.Activate()
                app.Selection.EndKey(Unit=app.wd('Story'))  # 跳到文档末尾
                app.Selection.Paste()
                member_doc.Close()
        master_doc.save()
        master_doc.Close(True)


class Document:
    """ 这个库写英文文档还不错。但不能做中文，字体会错乱。
    """

    def __init__(self, docx_file=None):
        """
        Args:
            docx_file:
                已有的word文件路径：打开
                还没创建的word文件路径：在个别功能需要的时候，会自动创建
                None：在临时文件夹生成一个默认的word文件
        """
        import docx
        # pip install python-docx

        if docx_file is None:
            self.docx_file = File(..., Dir.TEMP, suffix='.docx')
        else:
            self.docx_file = File(docx_file)
        if self.docx_file:
            self.doc = docx.Document(str(docx_file))
        else:
            self.doc = docx.Document()

    def write(self):
        Dir(self.docx_file.parent).ensure_dir()
        self.doc.save(str(self.docx_file))

    def to_pdf(self, pdf_file=None):
        self.write()
        pdf_file = DocxTools.to_pdf(self.docx_file, pdf_file)
        return pdf_file

    def to_fitzdoc(self):
        """ 获得 fitz的pdf文档对象
        :return: FitzDoc对象
        """
        from pyxllib.file.pdflib import FitzDoc
        pdf_file = self.to_pdf()
        doc = FitzDoc(pdf_file)
        return doc

    def to_images(self, file_fmt='{filestem}_{number}.png', *args, scale=1, **kwargs):
        doc = self.to_fitzdoc()
        files = doc.to_images(doc.src_file.parent, file_fmt, *args, scale=scale, **kwargs)
        return files

    def browser(self):
        """ 转pdf，使用浏览器的查看效果
        """
        pdf_file = self.to_pdf()
        browser(pdf_file)

    def display(self):
        """ 转图片，使用jupyter环境的查看效果
        """
        from pyxllib.cv.expert import xlpil
        xlpil.enchant()

        # 转图片，并且裁剪，加边框输出
        doc = self.to_fitzdoc()
        for i in range(doc.page_count):
            page = doc.load_page(i)
            print('= ' * 10 + f' {page} ' + '= ' * 10)
            img = page.get_pil_image()
            img.trim(border=5).plot_border().display()
            del page

    def to_labelmes(self, dst_dir=None, file_fmt='{filestem}_{number}.png', *, views=(0, 0, 1, 0), scale=1,
                    advance=False, indent=None):
        """ 转labelme格式查看

        本质是把docx转成pdf，利用pdf的解析生成labelme格式的标准框查看

        :param views: 详见to_labelmes的描述
            各位依次代表是否显示对应细粒度的标注：blocks、lines、spans、chars
        :param bool|dict advance: 是否开启“高级”功能，开启后能获得下划线等属性，但速度会慢很多
            源生的fitz pdf解析是处理不了下划线的，开启高级功能后，有办法通过特殊手段实现下划线的解析
            默认会修正目前已知的下划线、颜色偏差问题
            dict类型：以后功能多了，可能要支持自定义搭配，比如只复原unberline，但不管颜色偏差
        """
        from pyxllib.data.labelme import LabelmeDict

        # 1 转成图片，及json标注
        doc = self.to_fitzdoc()
        imfiles = doc.to_images(dst_dir, file_fmt, scale=scale)

        # 2 高级功能
        def is_color(x):
            return x and sum(x)

        def to_labelmes_advance():
            m = 50  # 匹配run时，上文关联的文字长度，越长越严格

            # 1 将带有下划线的run对象，使用特殊的hash规则存储起来
            content = []  # 使用已遍历到的文本内容作为hash值
            elements = {}
            for p in self.paragraphs:
                for r in p.runs:
                    # 要去掉空格，不然可能对不上。试过strip不行。分段会不太一样，还是把所有空格删了靠谱。
                    content.append(re.sub(r'\s+', '', r.text))
                    if r.underline or is_color(r.font.color.rgb):  # 对有下划线、颜色的对象进行存储
                        # print(r.text + ',', r.underline, r.font.color.rgb, ''.join(content))
                        etag = get_etag(''.join(content)[-m:])  # 全部字符都判断的话太严格了，所以减小区间~
                        elements[etag] = r

            # 2 检查json标注中为span的，其哈希规则是否有对应，则要单独设置扩展属性
            content = ''
            for i, file in enumerate(imfiles):
                page = doc.load_page(i)
                lmdict = LabelmeDict.gen_data(file)
                lmdict['shapes'] = page.get_labelme_shapes('dict', views=views, scale=scale)

                for sp in lmdict['shapes']:
                    attrs = DictTool.json_loads(sp['label'], 'label')
                    if attrs['category_name'] == 'span':
                        content += re.sub(r'\s+', '', attrs['text'])
                        etag = get_etag(content[-m:])
                        # print(content)
                        if etag in elements:
                            # print(content)
                            r = elements[etag]  # 对应的原run对象
                            attrs = DictTool.json_loads(sp['label'])
                            x = r.underline
                            if x:
                                attrs['underline'] = int(x)
                            x = r.font.color.rgb
                            if is_color(x):
                                attrs['color'] = list(x)
                            sp['label'] = json.dumps(attrs)
                file.with_suffix('.json').write(lmdict, indent=indent)

        # 3 获得json
        if advance:
            to_labelmes_advance()
        else:
            doc.to_labelmes(imfiles, views=views, scale=scale)

    def __getattr__(self, item):
        # 属性：
        # core_properties
        # element
        # inline_shapes
        # paragraphs
        # part
        # sections
        # settings
        # styles
        # tables
        # 方法：
        # add_heading
        # add_page_break
        # add_paragraph
        # add_picture
        # add_section
        # add_table
        # save
        return getattr(self.doc, item)


def __win32_word():
    """ 使用win32com调用word

    vba的文档：示例代码更多，vba语法也更熟悉，但显示的功能更不全
        https://docs.microsoft.com/en-us/office/vba/api/word.saveas2
    .net的文档：功能显示更全，应该是所有COM接口都有但示例代码更少、更不熟系
        https://docs.microsoft.com/en-us/dotnet/api/microsoft.office.interop.word.documentclass.saveas2?view=word-pia
    """
    pass


class EnchantWin32WordApplication(EnchantBase):
    @classmethod
    @run_once
    def enchant(cls, app, recursion_enchant=False):
        """
        :param app: win32的类是临时生成的，需要给一个参考对象，才方便type(word)算出类型
        :param recursion_enchant: 是否递归，对目前有的各种子类扩展功能都绑定上
            默认关闭，如果影响到性能，可以关闭，后面运行中需要时手动设定enchant
            开启，能方便业务层开发

            之前有想过可以生成doc里的时候再enchant这些对象，但如果是批量处理脚本，每次建立doc都判断我觉得也麻烦
            长痛不如短痛，建立app的时候就把所有对象enchant更方便
        """
        # app
        _cls = type(app)
        names = cls.check_enchant_names([_cls], ignore_case=True)
        exclude_names = {'get_app'}
        cls._enchant(_cls, names - exclude_names)

        if recursion_enchant:
            # 建一个临时文件，把各种需要绑定的对象都生成绑定一遍
            # 确保初始化稍微慢点，但后面就方便了
            doc = app.Documents.Add()
            EnchantWin32WordDocument.enchant(doc)

            doc.Activate()
            rng = doc.Range()  # 全空的文档，有区间[0,1)
            EnchantWin32WordRange.enchant(rng)

            doc.Hyperlinks.Add(rng, 'url')  # 因为全空，这里会自动生成对应的明文url
            EnchantWin32WordHyperlink.enchant(doc.Hyperlinks(1))

            # 处理完关闭文档，不用保存
            doc.Close(False)

    @classmethod
    def get_app(cls, app=None, *, visible=None, display_alerts=0, recursion_enchant=True):
        """
        Args:
            app: 可以自定义在外部使用Dispatch、DispatchEx等形式给入已初始化好的app
            visible: 是否可见
            display_alerts: 是否关闭警告
            recursion_enchant: 是否递归执行enchant
        """
        # 1 get app
        name = 'WORD.APPLICATION'
        if app is None:
            try:
                app = win32.GetActiveObject(name)
            except pythoncom.com_error:
                pass
        if app is None:
            try:
                # 名称用大写，会比较兼容旧的word2013等版本
                # 尽量静态调度，才能获得 from win32com.client import constants 的常量
                app = win32.gencache.EnsureDispatch(name)
            except TypeError:
                # 实在不行，就用动态调度
                app = win32.dynamic.Dispatch(name)

        # 2 enchant
        cls.enchant(app, recursion_enchant=recursion_enchant)

        if visible is not None:
            app.Visible = visible
        if display_alerts is not None:
            app.DisplayAlerts = display_alerts  # 不警告

        return app

    @staticmethod
    def check_close(app, outfile):
        """ 检查是否有指定名称的文件被打开，将其关闭，避免new_doc等操作出现问题
        """
        outfile = File(outfile)
        for x in app.Documents:
            # 有可能文件来自onedrive，这里用safe_init更合理
            if File.safe_init(x.Name, x.Path) == outfile:
                x.Close()

    @staticmethod
    def open_doc(app, file_name):
        """ 打开已有的文件
        """
        doc = app.Documents.Open(str(file_name))
        return doc

    @staticmethod
    def new_doc(app, file=None):
        """ 创建一个新的文件
        Args:
            file: 文件路径
                空：新建一个doc，到时候保存会默认到临时文件夹
                不存在的文件名：新建对应的空文件
                已存在的文件名：重置、覆盖一个新的空文件

        使用该函数，会自动执行EnchantWin32WordDocument扩展。
        """
        if file is None:
            file = File(..., Dir.TEMP, suffix='.docx')
        else:
            file = File(file)

        doc = app.Documents.Add()  # 创建新的word文档
        doc.save(file)
        return doc

    @staticmethod
    def wd(app, name, part=None):
        """ 输入字符串名称，获得对应的常量值

        :param name: 必须省略前缀wd。这个函数叫wd，就是帮忙省略掉前缀wd的意思
        :param part: 特定组别的枚举值，可以输入特殊的name模式来取值
        """
        if part is None:
            return getattr(constants, 'wd' + name)
        else:
            raise ValueError


class EnchantWin32WordDocument(EnchantBase):
    @classmethod
    @run_once
    def enchant(cls, doc):
        _cls = type(doc)
        names = cls.check_enchant_names([_cls], ignore_case=True)
        propertys = {'n_page'}
        cls._enchant(_cls, propertys, EnchantCvt.staticmethod2property)
        cls._enchant(_cls, names - propertys)

    @staticmethod
    def save(doc, file_name=None, fmt=None, retain=False, **kwargs):
        """ 我自己简化的保存接口

        :param file_name: 保存到指定路径，如果带有后缀
        :param fmt: 毕竟是底层的com接口，不能做的太智能吧。连通过文件名后缀自动选择格式的功能都没有，要手动指定。
            为了方便，对这些进行智能自动处理，得到一个合理的save接口。
        :param retain: SaveAs2的机制：如果目标格式仍是word支持的，则doc会切换到目标文件。否则doc保留原文件对象。
            这里retain若打开，则会自动做切换，保留原文件对象
        :return: 跟retain有关，可能会"重置", outfile
            默认返回 outfile
            开启retain时，返回 outfile, doc
        """

        # 1 辅助函数
        def save_format(fmt):
            """ 枚举值映射，word保存类型的枚举

            >> _('.html')
            8
            >> _('Pdf')
            17
            >> _('wdFormatFilteredHTML')
            10
            """
            # 复杂格式可能无法完美支持所有功能。比如复杂的pdf无法使用SaveAs2实现，要用ExportAsFixedFormat。
            # 有些情况可能无法使用gencache，导致没有constants，所以默认可以直接映射到整数，避免使用到constants
            common = {'doc': 0,
                      'html': 8,
                      'txt': 2,
                      'docx': 16,
                      'pdf': 17}
            name = common.get(fmt.lower().lstrip('.'), fmt)
            if isinstance(name, int):
                return name
            else:
                return getattr(constants, 'wd' + name)

        # 2 确认要存储的文件格式
        if isinstance(fmt, str):
            fmt = fmt.lower().lstrip('.')
        elif file_name is not None:
            fmt = File(file_name).suffix[1:].lower()
        elif doc.Path:
            fmt = os.path.splitext(doc.Name)[1][1:].lower()
        else:
            fmt = 'docx'

        # 3 保存一份原始的文件路径
        origin_file = File(doc.Name, doc.Path) if doc.Path else None

        # 4 如果有指定保存文件路径
        if file_name is not None:
            outfile = File(file_name)
            if outfile.suffix[1:].lower() != fmt:
                # 已有文件名，但这里指定的fmt不同于原文件，则认为是要另存为一个同名的不同格式文件
                outfile = File(outfile.stem, outfile.parent, suffix=fmt)
            doc.SaveAs2(str(outfile), save_format(fmt), **kwargs)
        # 5 如果没指定保存文件路径
        else:
            if doc.Path:
                outfile = File(doc.Name, doc.Path, suffix='.' + fmt)
                doc.SaveAs2(str(outfile), save_format(outfile.suffix), **kwargs)
            else:
                etag = get_etag(doc.Content)
                outfile = File(etag, Dir.TEMP, suffix=fmt)
                doc.SaveAs2(str(outfile), save_format(fmt), **kwargs)

        # 6 是否恢复原doc
        cur_file = File(doc.Name, doc.Path)  # 当前文件不一定是目标文件f，如果是pdf等格式也不会切换过去
        if retain and origin_file and origin_file != cur_file:
            app = doc.Application
            doc.Close()
            doc = app.open_doc(origin_file)

        # 7 返回值
        if retain:
            return outfile, doc
        else:
            return outfile

    # 先暂时不开启 doc.chars
    # @staticmethod
    # def chars(doc):
    #     return doc.Range().chars

    @staticmethod
    def n_page(doc):
        return doc.ActiveWindow.Panes(1).Pages.Count

    @staticmethod
    def browser(doc, file_name=None, fmt='html', retain=False):
        """ 这个函数可能会导致原doc指向对象被销毁，建议要不追返回值doc继续使用
        """
        res = doc.save(file_name, fmt, retain=retain)

        if retain:
            outfile, doc = res
        else:
            outfile = res

        browser(outfile)
        return doc

    @staticmethod
    def add_section_size(doc, factor=1):
        """ 增加每节长度的标记
        一般在这里计算比在html计算方便
        """
        from humanfriendly import format_size

        n = doc.Paragraphs.Count
        style_names, text_lens = [], []
        for p in doc.Paragraphs:
            style_names.append(str(p.Style))
            text_lens.append(strwidth(p.Range.Text))

        for i, p in enumerate(doc.Paragraphs):
            name = style_names[i]
            if name.startswith('标题'):
                cumulate_size = 0
                for j in range(i + 1, n):
                    if style_names[j] != name:
                        cumulate_size += text_lens[j - 1]
                    else:
                        break
                if cumulate_size:
                    size = format_size(cumulate_size * factor).replace(' ', '').replace('bytes', 'B')
                    r = p.Range
                    doc.Range(r.Start, r.End - 1).InsertAfter(f'，{size}')

    @staticmethod
    def outline_demote(doc, demote_level):
        """ 标题降级，降低level层 """
        for p in doc.Paragraphs:
            p.Range.demote(demote_level)

    @staticmethod
    def set_style(doc, obj, name):
        """ 给Paragraph、Range等按名称设置样式

        :param obj: 当前doc下某个含有Style成员属性的子对象
        :param name: 样式名称
        """
        setattr(obj, 'Style', doc.Styles(name))

    @staticmethod
    def add_paragraph(doc, text='', style=None):
        """ 自定义的插入段落

        默认的插入比较麻烦，新建段落、插入文本、设置格式要多步实现，这里封装支持在一步进行多种操作
        """
        p = doc.Paragraphs.Add()
        if text:
            p.Range.InsertBefore(text)
        if style:
            p.Style = doc.Styles(style)
        return p


class EnchantWin32WordRange(EnchantBase):
    """ range是以下标0开始，左闭右开的区间

    当一个区间出现混合属性，比如有的有加粗，有的没加粗时，标记值为 app.wd('Undefined') 9999999
    vba的True是值-1，False是值0
    """

    @classmethod
    @run_once
    def enchant(cls, rng):
        _cls = type(rng)
        names = cls.check_enchant_names([_cls], ignore_case=True)
        propertys = {'chars'}
        cls._enchant(_cls, propertys, EnchantCvt.staticmethod2property)
        cls._enchant(_cls, names - propertys)

    @staticmethod
    def set_hyperlink(rng, url):
        """ 给当前rng添加超链接
        """
        doc = rng.Parent
        doc.Hyperlinks.Add(rng, url)

    @staticmethod
    def chars(rng):
        # 有特殊换行，ch.Text可能会得到 '\r\x07'，为了位置对应，只记录一个字符
        return ''.join([ch.Text[0] for ch in rng.Characters])

    @staticmethod
    def char_range(rng, start=0, end=None):
        """ 定位rng中的子range对象，这里是以可见字符Characters计数的

        :param start: 下标类似切片的规则
        :param end: 见start描述，允许越界，允许负数
            默认不输入表示匹配到末尾
        """
        n = rng.Characters.Count
        if end is None or end > n:
            end = n
        elif end < 0:
            end = n + end
        start_idx, end_idx = rng.Characters(start + 1).Start, rng.Characters(end).End
        return rng.Document.Range(start_idx, end_idx)

    @staticmethod
    def shifting(rng, left=0, right=0):
        """ range左右两边增加偏移量，返回重定位的rng

        常用语段落定位，要在段落末尾增加内容时
        >> rng2 = p.Range.shifting(right=-1)
        """
        return rng.Document.Range(rng.Start + left, rng.End + right)

    @staticmethod
    def demote(rng, demote_level):
        """ 标题降级，降低level层 """
        name = rng.Style.NameLocal  # 获得样式名称
        m = re.match(r'标题 (\d)$', name)
        if m:
            lvl = int(m.group(1))
            new_lvl = lvl + demote_level
            new_style = f'标题 {new_lvl}' if new_lvl < 10 else '正文'
            rng.Style = rng.Parent.Styles(new_style)

    @staticmethod
    def set_font(rng, font_fmt):
        """ 设置字体各种格式

        :param dict|str font_fmt:
            dict，定制格式
                布尔类型：Bold、Italic、Subscript、Superscript
                可布尔的值类型：Underline
                    支持的格式见：https://docs.microsoft.com/en-us/office/vba/api/word.wdunderline
                值类型：Name、Size、Color、UnderlineColor
            str，使用现有样式名
        """
        if isinstance(font_fmt, dict):
            font = rng.Font
            for k, v in font_fmt.items():
                setattr(font, k, v)
        elif isinstance(font_fmt, str):
            rng.Style = rng.Parent.Styles(font_fmt)
        else:
            raise ValueError

    @staticmethod
    def insert_before(rng, text, font_fmt=None):
        """ 对原InsertBefore的功能封装

        :return: 增加返回值，是新插入内容的range定位
        """
        start1, end1 = rng.Start, rng.End
        rng.InsertBefore(text)
        bias = rng.End - end1  # 新插入的内容大小
        new_rng = rng.Document.Range(start1, start1 + bias)
        if font_fmt:
            new_rng.set_font(font_fmt)
        return new_rng

    @staticmethod
    def insert_after(rng, text, font_fmt=None):
        """ 同insert_before，是InsertAfter的重封装
        """
        # 1
        start1, end1 = rng.Start, rng.End

        # 2 往后插入，会排除\r情况
        doc = rng.Document
        ch = doc.Range(end1 - 1, end1).Text
        if ch == '\r':
            end1 -= 1
            rng = doc.Range(start1, end1)

        # 3
        rng.InsertAfter(text)
        bias = rng.End - end1
        new_rng = rng.Document.Range(end1, end1 + bias)
        if font_fmt:
            new_rng.set_font(font_fmt)
        return new_rng


class EnchantWin32WordHyperlink(EnchantBase):
    @classmethod
    @run_once
    def enchant(cls, link):
        _cls = type(link)
        names = cls.check_enchant_names([_cls], ignore_case=True)
        propertys = {'netloc', 'name'}
        cls._enchant(_cls, propertys, EnchantCvt.staticmethod2property)
        cls._enchant(_cls, names - propertys)

    @staticmethod
    def netloc(link):
        from urllib.parse import urlparse
        linkp = urlparse(link.Name)  # 链接格式解析
        # netloc = linkp.netloc or Path(linkp.path).name
        netloc = linkp.netloc or linkp.scheme  # 可能是本地文件，此时记录其所在磁盘
        return netloc

    @staticmethod
    def name(link):
        """ 这个是转成明文的完整链接，如果要编码过的，可以取link.Name """
        from urllib.parse import unquote
        return unquote(link.Name)


def rebuild_document_by_word(fmt='html', translate=False, navigation=False, visible=False, quit=None):
    """ 将剪切板的内容粘贴到word重新排版，再转成fmt格式的文档，用浏览器打开

    这个功能只能在windows平台使用，并且必须要安装有Word软件。

    一般用于英文网站，生成双语阅读的模板，再调用谷歌翻译。
    生成的文档如果有需要，一般是手动复制整理到另一个docx文件中。

    Args:
        fmt: 输出文件类型
            常见的可以简写：html、pdf、txt
            其他特殊需求可以用word原变量名：wdFormatDocument
        visible: 是否展示运行过程，如果不展示，默认最后会close文档
        quit: 运行完是否退出应用
        translate: html专用业务功能，表示是否对p拷贝一份notranslate的对象，用于谷歌翻译双语对照
        navigation: 是否增加导航栏
            注意，使用导航栏后，页面就无法直接使用谷歌翻译了
            但可以自己进入_content文件，使用谷歌翻译处理，自覆盖保存
            然后再回到_index文件，刷新即可
    """
    import pyperclip
    from pyxllib.text.xmllib import BeautifulSoup, html_bitran_template, MakeHtmlNavigation

    # 1 保存的临时文件名采用etag
    f = File(get_etag(pyperclip.paste()), Dir.TEMP, suffix=fmt)
    app = EnchantWin32WordApplication.get_app(visible=visible)
    app.check_close(f)
    doc = app.new_doc(f)
    doc.Activate()
    app.Selection.Paste()

    # 2 如果需要，也可以在这个阶段，插入word自动化的操作，而不是后续在html层面操作
    # 统计每节内容长度，每个字母1B，每个汉字2B
    doc.add_section_size()
    file = doc.save(f, fmt)
    doc.Close()

    # 3 html格式扩展功能
    if fmt == 'html':
        # 3.1 默认扩展功能
        s = file.read(encoding='gbk')
        # s = s.replace('\xa0', '')  # 不知道这样去除\xa0行不行，等下次遇到尝试
        bs = BeautifulSoup(s, 'lxml')
        bs.head_add_number()  # 给标题加上编号
        # bs.head_add_size()  # 显示每节内容长短
        content = str(bs)

        # TODO 识别微信、pydoc，然后做一些自动化清理？
        # TODO 过度缩进问题？

        # 3.2 双语对照阅读
        if translate:
            # word生成的html固定是gbk编码
            content = html_bitran_template(content)

        # 原文是gbk，但到谷歌默认是utf8，所以改一改
        # 这样改后问题是word可能又反而有问题了，不过word本来只是跳板，并不是要用word编辑html
        file.write(content, encoding='utf8')

        # 3.3 导航栏功能
        # 作为临时使用可以开，如果要复制到word，并没有必要
        if navigation:
            file = MakeHtmlNavigation.from_file(file, encoding='utf8', number=False, text_catalogue=False)

    if quit:
        app.Quit()

    return file



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import subprocess

__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '..')))

os.environ["FLAGS_allocator_strategy"] = 'auto_growth'
import cv2
import numpy as np
import time
import logging

from ppocr.utils.utility import get_image_file_list, check_and_read_gif
from ppocr.utils.logging import get_logger
from tools.infer.predict_system import TextSystem
from ppstructure.table.predict_table import TableSystem, to_excel
from ppstructure.utility import parse_args, draw_structure_result

logger = get_logger()


class OCRSystem(object):
    def __init__(self, args):
        self.mode = args.mode
        if self.mode == 'structure':
            import layoutparser as lp
            # args.det_limit_type = 'resize_long'
            args.drop_score = 0
            if not args.show_log:
                logger.setLevel(logging.INFO)
            self.text_system = TextSystem(args)
            self.table_system = TableSystem(args,
                                            self.text_system.text_detector,
                                            self.text_system.text_recognizer)

            config_path = None
            model_path = None
            if os.path.isdir(args.layout_path_model):
                model_path = args.layout_path_model
            else:
                config_path = args.layout_path_model
            self.table_layout = lp.PaddleDetectionLayoutModel(
                config_path=config_path,
                model_path=model_path,
                threshold=0.5,
                enable_mkldnn=args.enable_mkldnn,
                enforce_cpu=not args.use_gpu,
                thread_num=args.cpu_threads)
            self.use_angle_cls = args.use_angle_cls
            self.drop_score = args.drop_score
        elif self.mode == 'vqa':
            from ppstructure.vqa.infer_ser_e2e import SerPredictor, draw_ser_results
            self.vqa_engine = SerPredictor(args)

    def __call__(self, img):
        if self.mode == 'structure':
            ori_im = img.copy()
            layout_res = self.table_layout.detect(img[..., ::-1])
            res_list = []
            for region in layout_res:
                x1, y1, x2, y2 = region.coordinates
                x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
                roi_img = ori_im[y1:y2, x1:x2, :]
                if region.type == 'Table':
                    res = self.table_system(roi_img)
                else:
                    filter_boxes, filter_rec_res = self.text_system(roi_img)
                    filter_boxes = [x + [x1, y1] for x in filter_boxes]
                    filter_boxes = [
                        x.reshape(-1).tolist() for x in filter_boxes
                    ]
                    # remove style char
                    style_token = [
                        '<strike>', '<strike>', '<sup>', '</sub>', '<b>',
                        '</b>', '<sub>', '</sup>', '<overline>', '</overline>',
                        '<underline>', '</underline>', '<i>', '</i>'
                    ]
                    filter_rec_res_tmp = []
                    for rec_res in filter_rec_res:
                        rec_str, rec_conf = rec_res
                        for token in style_token:
                            if token in rec_str:
                                rec_str = rec_str.replace(token, '')
                        filter_rec_res_tmp.append((rec_str, rec_conf))
                    res = (filter_boxes, filter_rec_res_tmp)
                res_list.append({
                    'type': region.type,
                    'bbox': [x1, y1, x2, y2],
                    'img': roi_img,
                    'res': res
                })
        elif self.mode == 'vqa':
            res_list, _ = self.vqa_engine(img)
        return res_list


def save_structure_res(res, save_folder, img_name):
    excel_save_folder = os.path.join(save_folder, img_name)
    os.makedirs(excel_save_folder, exist_ok=True)
    # save res
    with open(
            os.path.join(excel_save_folder, 'res.txt'), 'w',
            encoding='utf8') as f:
        for region in res:
            if region['type'] == 'Table':
                excel_path = os.path.join(excel_save_folder,
                                          '{}.xlsx'.format(region['bbox']))
                to_excel(region['res'], excel_path)
            if region['type'] == 'Figure':
                roi_img = region['img']
                img_path = os.path.join(excel_save_folder,
                                        '{}.jpg'.format(region['bbox']))
                cv2.imwrite(img_path, roi_img)
            else:
                for box, rec_res in zip(region['res'][0], region['res'][1]):
                    f.write('{}\t{}\n'.format(
                        np.array(box).reshape(-1).tolist(), rec_res))


def main(args):
    image_file_list = get_image_file_list(args.image_dir)
    image_file_list = image_file_list
    image_file_list = image_file_list[args.process_id::args.total_process_num]

    structure_sys = OCRSystem(args)
    img_num = len(image_file_list)
    save_folder = os.path.join(args.output, structure_sys.mode)
    os.makedirs(save_folder, exist_ok=True)

    for i, image_file in enumerate(image_file_list):
        logger.info("[{}/{}] {}".format(i, img_num, image_file))
        img, flag = check_and_read_gif(image_file)
        img_name = os.path.basename(image_file).split('.')[0]

        if not flag:
            img = cv2.imread(image_file)
        if img is None:
            logger.error("error in loading image:{}".format(image_file))
            continue
        starttime = time.time()
        res = structure_sys(img)

        if structure_sys.mode == 'structure':
            save_structure_res(res, save_folder, img_name)
            draw_img = draw_structure_result(img, res, args.vis_font_path)
            img_save_path = os.path.join(save_folder, img_name, 'show.jpg')
        elif structure_sys.mode == 'vqa':
            draw_img = draw_ser_results(img, res, args.vis_font_path)
            img_save_path = os.path.join(save_folder, img_name + '.jpg')
        cv2.imwrite(img_save_path, draw_img)
        logger.info('result save to {}'.format(img_save_path))
        elapse = time.time() - starttime
        logger.info("Predict time : {:.3f}s".format(elapse))


if __name__ == "__main__":
    args = parse_args()
    if args.use_mp:
        p_list = []
        total_process_num = args.total_process_num
        for process_id in range(total_process_num):
            cmd = [sys.executable, "-u"] + sys.argv + [
                "--process_id={}".format(process_id),
                "--use_mp={}".format(False)
            ]
            p = subprocess.Popen(cmd, stdout=sys.stdout, stderr=sys.stdout)
            p_list.append(p)
        for p in p_list:
            p.wait()
    else:
        main(args)



# copyright (c) 2021 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import paddle
from paddle import nn
from .ace_loss import ACELoss
from .center_loss import CenterLoss
from .rec_ctc_loss import CTCLoss


class EnhancedCTCLoss(nn.Layer):
    def __init__(self,
                 use_focal_loss=False,
                 use_ace_loss=False,
                 ace_loss_weight=0.1,
                 use_center_loss=False,
                 center_loss_weight=0.05,
                 num_classes=6625,
                 feat_dim=96,
                 init_center=False,
                 center_file_path=None,
                 **kwargs):
        super(EnhancedCTCLoss, self).__init__()
        self.ctc_loss_func = CTCLoss(use_focal_loss=use_focal_loss)

        self.use_ace_loss = False
        if use_ace_loss:
            self.use_ace_loss = use_ace_loss
            self.ace_loss_func = ACELoss()
            self.ace_loss_weight = ace_loss_weight

        self.use_center_loss = False
        if use_center_loss:
            self.use_center_loss = use_center_loss
            self.center_loss_func = CenterLoss(
                num_classes=num_classes,
                feat_dim=feat_dim,
                init_center=init_center,
                center_file_path=center_file_path)
            self.center_loss_weight = center_loss_weight

    def __call__(self, predicts, batch):
        loss = self.ctc_loss_func(predicts, batch)["loss"]

        if self.use_center_loss:
            center_loss = self.center_loss_func(
                predicts, batch)["loss_center"] * self.center_loss_weight
            loss = loss + center_loss

        if self.use_ace_loss:
            ace_loss = self.ace_loss_func(
                predicts, batch)["loss_ace"] * self.ace_loss_weight
            loss = loss + ace_loss

        return {'enhanced_ctc_loss': loss}



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import paddle
from paddle import nn


class SRNLoss(nn.Layer):
    def __init__(self, **kwargs):
        super(SRNLoss, self).__init__()
        self.loss_func = paddle.nn.loss.CrossEntropyLoss(reduction="sum")

    def forward(self, predicts, batch):
        predict = predicts['predict']
        word_predict = predicts['word_out']
        gsrm_predict = predicts['gsrm_out']
        label = batch[1]

        casted_label = paddle.cast(x=label, dtype='int64')
        casted_label = paddle.reshape(x=casted_label, shape=[-1, 1])

        cost_word = self.loss_func(word_predict, label=casted_label)
        cost_gsrm = self.loss_func(gsrm_predict, label=casted_label)
        cost_vsfd = self.loss_func(predict, label=casted_label)

        cost_word = paddle.reshape(x=paddle.sum(cost_word), shape=[1])
        cost_gsrm = paddle.reshape(x=paddle.sum(cost_gsrm), shape=[1])
        cost_vsfd = paddle.reshape(x=paddle.sum(cost_vsfd), shape=[1])

        sum_cost = cost_word * 3.0 + cost_vsfd + cost_gsrm * 0.15

        return {'loss': sum_cost, 'word_loss': cost_word, 'img_loss': cost_vsfd}



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import numpy as np
import os
import random
import traceback
from paddle.io import Dataset
from .imaug import transform, create_operators

import json

from pyxllib.xl import run_once, XlPath

__all__ = ['SimpleDataSet', 'XlSimpleDataSet']


class SimpleDataSet(Dataset):
    """ paddleocr 源生的基础数据格式
    每张图的标注压缩在一个总的txt文件里
    """

    def __init__(self, config, mode, logger, seed=None):
        super(SimpleDataSet, self).__init__()
        self.logger = logger
        self.mode = mode.lower()

        # 这里能取到全局的配置信息
        global_config = config['Global']
        dataset_config = config[mode]['dataset']
        loader_config = config[mode]['loader']

        # label文件里图片路径和json之间的分隔符
        self.delimiter = dataset_config.get('delimiter', '\t')

        label_file_list = dataset_config.pop('label_file_list')
        data_source_num = len(label_file_list)
        ratio_list = dataset_config.get("ratio_list", [1.0])
        if isinstance(ratio_list, (float, int)):
            ratio_list = [float(ratio_list)] * int(data_source_num)

        assert len(
            ratio_list
        ) == data_source_num, "The length of ratio_list should be the same as the file_list."
        self.data_dir = dataset_config['data_dir']
        self.do_shuffle = loader_config['shuffle']

        self.seed = seed
        logger.info("Initialize indexs of datasets:%s" % label_file_list)
        self.data_lines = self.get_image_info_list(label_file_list, ratio_list)
        self.data_idx_order_list = list(range(len(self.data_lines)))
        # shuffle好像在dataset和dataloader会重复操作，虽然其实也没什么事~~
        if self.mode == "train" and self.do_shuffle:
            self.shuffle_data_random()
        # （读取图片）数据增广操作器
        self.ops = create_operators(dataset_config['transforms'], global_config)

    def get_image_info_list(self, file_list, ratio_list):
        """ 从多个文件按比例随机抽样获取样本 """
        if isinstance(file_list, str):
            file_list = [file_list]
        data_lines = []
        for idx, file in enumerate(file_list):
            with open(file, "rb") as f:
                lines = f.readlines()
                if self.mode == "train" or ratio_list[idx] < 1.0:
                    random.seed(self.seed)
                    lines = random.sample(lines,
                                          round(len(lines) * ratio_list[idx]))
                data_lines.extend(lines)
        return data_lines

    def shuffle_data_random(self):
        random.seed(self.seed)
        random.shuffle(self.data_lines)
        return

    def get_ext_data(self):
        """ 是否要添加其他图片及数量

        猜测是用于mixup、cropmix等场合的数据增广，除了当前图，能随机获取其他来源图片，做综合处理
        其它来源的图，会调用前2个opts，读取图片，解析标签
        """
        ext_data_num = 0
        for op in self.ops:
            if hasattr(op, 'ext_data_num'):
                ext_data_num = getattr(op, 'ext_data_num')
                break
        load_data_ops = self.ops[:2]
        ext_data = []

        while len(ext_data) < ext_data_num:
            # 随机从中抽一个样本
            file_idx = self.data_idx_order_list[np.random.randint(self.__len__(
            ))]
            data_line = self.data_lines[file_idx]
            data_line = data_line.decode('utf-8')
            substr = data_line.strip("\n").split(self.delimiter)
            file_name = substr[0]
            label = substr[1]
            img_path = os.path.join(self.data_dir, file_name)
            data = {'img_path': img_path, 'label': label}
            if not os.path.exists(img_path):
                continue
            with open(data['img_path'], 'rb') as f:
                img = f.read()
                data['image'] = img
            data = transform(data, load_data_ops)

            if data is None or data['polys'].shape[1]!=4:
                continue
            ext_data.append(data)
        return ext_data

    def __getitem__(self, idx):
        file_idx = self.data_idx_order_list[idx]
        data_line = self.data_lines[file_idx]
        # print(data_line.split()[0])
        try:
            data_line = data_line.decode('utf-8')
            substr = data_line.strip("\n").split(self.delimiter)
            file_name = substr[0]
            label = substr[1]
            img_path = os.path.join(self.data_dir, file_name)
            data = {'img_path': img_path, 'label': label}
            if not os.path.exists(img_path):
                raise Exception("{} does not exist!".format(img_path))
            # 无论是图片还是任何文件，都是统一先读成bytes了，然后交由transform的配置实现。
            #	其中DecodeImage又可以解析读取图片数据。
            with open(data['img_path'], 'rb') as f:
                img = f.read()
                data['image'] = img
            data['ext_data'] = self.get_ext_data()
            if data['label'] == '[]':  # 没有文本的图片
                data['label'] = '[{"transcription": "###", "points": [[0, 0], [1, 0], [1, 1], [0, 1]]}]'
            outs = transform(data, self.ops)
        except Exception as e:
            self.logger.error(
                "When parsing line {}, error happened with msg: {}".format(
                    data_line, e))
            outs = None
        if outs is None:
            # 如果遇到解析出错的数据，训练阶段会随机取另一个图片代替。
            #	eval阶段则直接取下一张有效图片代替。
            # during evaluation, we should fix the idx to get same results for many times of evaluation.
            rnd_idx = np.random.randint(self.__len__(
            )) if self.mode == "train" else (idx + 1) % self.__len__()
            return self.__getitem__(rnd_idx)
        return outs

    def __len__(self):
        return len(self.data_idx_order_list)


class SimpleDataSetExt(SimpleDataSet):
    """ 自定义的数据结构类，支持输入标注文件所在目录来初始化

    这里的 __init__、get_image_info_list 设计了一套特殊的输入范式
    """

    def __init__(self, config, mode, logger, seed=None):
        self.logger = logger
        self.mode = mode.lower()

        # 这里能取到全局的配置信息
        global_config = config['Global']
        dataset_config = config[mode]['dataset']
        loader_config = config[mode]['loader']

        # label文件里图片路径和json之间的分隔符
        self.delimiter = dataset_config.get('delimiter', '\t')

        self.data_dir = dataset_config['data_dir']
        self.do_shuffle = loader_config['shuffle']

        self.seed = seed
        data_list = dataset_config.get('data_list', [])
        logger.info("Initialize indexs of datasets:%s" % data_list)
        self.data_lines = self.get_image_info_list(data_list)
        self.data_idx_order_list = list(range(len(self.data_lines)))
        # shuffle好像在dataset和dataloader会重复操作，虽然其实也没什么事~~
        if self.mode == "train" and self.do_shuffle:
            self.shuffle_data_random()
        # （读取图片）数据增广操作器
        self.ops = create_operators(dataset_config['transforms'], global_config)

    def get_image_info_list(self, data_list):
        """ 从标注文件所在目录获取每张图的标注信息

        :return: list data_lines，每一行有两列，第1列是图片数据相对data_dir的路径，\t隔开，第2列是json标注数据
        """
        if isinstance(data_list, dict):
            data_list = [data_list]

        data_lines = []

        for idx, cfg in enumerate(data_list):
            add_lines = [x.encode('utf8') for x in self.get_image_info(cfg)]
            data_lines.extend(add_lines)

        return data_lines

    def get_image_info(self, cfg):
        raise NotImplementedError


class XlSimpleDataSet(SimpleDataSetExt):
    """ 支持直接配置原始的icdar2015数据格式
    """

    def get_image_info(self, cfg):
        # 我的XlSimpleDataSet，type是必填字段
        t = cfg.pop('type') if ('type' in cfg) else ''
        func = getattr(self, 'from_' + t, None)
        if func:
            return func(**cfg)
        else:
            raise TypeError('指定数据集格式不存在')

    def __repr__(self):
        """ 跟runonce有关，需要用这个构造类字符串，判断参数是否重复 """
        args = [self.data_dir, self.seed]
        return 'XlSimpleDataSet(' + ','.join(map(str, args)) + ')'

    def _select_ratio(self, data, ratio):
        """ 随机筛选给定的data数组数据

        方便有些数据没有物理地址划分训练、验证集的，可以通过ratio直接设置，目前只支持一个数值ratio，后续也可以考虑支持list，更灵活的截选策略
        """
        random.seed(4101)  # 我这里是要用固定策略拆分数据，不用self.seed
        random.shuffle(data)
        n = len(data)
        if isinstance(ratio, float):
            if ratio > 0:
                data = data[:int(ratio * n)]
            elif ratio < 0:
                data = data[int(ratio * n):]
        elif isinstance(ratio, list):
            left, right = ratio  # 这个ratio是每个类别的模板分开处理的
            data = data[int(left * n):int(right * n)]
        return data

    @run_once('str')  # 这个标注格式是固定的，不用每次重复生成，可以使用run_once限定
    def from_icdar2015(self, img_dir, label_dir, ratio=None):
        data_dir = XlPath(self.data_dir)
        img_dir = data_dir / img_dir
        label_dir = data_dir / label_dir

        data_lines = []

        txt_files = list(label_dir.glob('*.txt'))
        if ratio is not None:
            txt_files = self._select_ratio(txt_files, ratio)

        def label2json(content):
            """ 单个图的content标注内容转为json格式 """
            label = []
            for line in content.splitlines():
                tmp = line.split(',')
                points = tmp[:8]
                s = []
                for i in range(0, len(points), 2):
                    b = points[i:i + 2]
                    b = [int(t) for t in b]
                    s.append(b)
                result = {"transcription": tmp[8], "points": s}
                label.append(result)
            return label

        for f in txt_files:
            # stem[3:]是去掉标注文件名多出的'gt_'的前缀
            impath = (img_dir / (f.stem[3:] + '.jpg')).relative_to(data_dir).as_posix()
            # icdar的标注文件，有的是utf8，有的是utf-8-sig，这里使用我的自动识别功能
            json_label = label2json(f.read_text(encoding=None))
            label = json.dumps(json_label, ensure_ascii=False)
            data_lines.append(('\t'.join([impath, label])))

        return data_lines

    @run_once('str')
    def from_refineAgree(self, img_dir, json_dir, label_file):
        """ 只需要输入根目录 """
        from pyxllib.data.labelme import LabelmeDict

        data_dir = XlPath(self.data_dir)
        img_dir = data_dir / img_dir
        json_dir = data_dir / json_dir
        label_file = data_dir / label_file

        def labelme2json(d):
            """ labelme的json转为paddle的json标注 """
            label = []
            shapes = d['shapes']
            for sp in shapes:
                msg = json.loads(sp['label'])
                if msg['type'] != '印刷体':
                    continue
                result = {"transcription": msg['text'],
                          "points": LabelmeDict.to_quad_pts(sp)}
                label.append(result)
            return label

        data_lines = []
        sample_list = label_file.read_text().splitlines()
        for x in sample_list:
            if not x: continue  # 忽略空行
            impath = (img_dir / (x + '.jpg')).relative_to(data_dir).as_posix()
            f = json_dir / (x + '.json')
            json_label = labelme2json(f.read_json())
            label = json.dumps(json_label, ensure_ascii=False)
            data_lines.append(('\t'.join([impath, label])))

        return data_lines

    @run_once('str')
    def from_labelme_det(self, sub_data_dir='.', ratio=None, transcription_field=None):
        """ 读取sub_data_dir目录（含子目录）下所有的json文件为标注文件

        :param transcription_field: 对于检测任务这个值一般没什么用，主要是一些特殊数据，标记"#"的会记为难样本，跳过不检测
            默认不设置取sp['label']为文本值
            若设置，则按字典解析label并取对应名称的键值

        json1: labelme的json标注文件
        json2: paddle的SimpleDataSet要传入的json格式
        """
        from pyxllib.data.labelme import LabelmeDict

        data_dir = XlPath(self.data_dir)
        sub_data_dir = data_dir / sub_data_dir
        data_lines = []

        def json1_to_json2(d):
            res = []
            shapes = d['shapes']
            for sp in shapes:
                label = sp['label']
                if transcription_field:
                    msg = json.loads(sp['label'])
                    label = msg[transcription_field]
                result = {"transcription": label,
                          "points": LabelmeDict.to_quad_pts(sp)}
                res.append(result)
            return res

        json1_files = list(sub_data_dir.rglob('*.json'))
        if ratio is not None:
            json1_files = self._select_ratio(json1_files, ratio)

        for json1_file in json1_files:
            data = json1_file.read_json()
            img_file = json1_file.parent / data['imagePath']  # 不确定关联的图片格式，所以直接从labelme里取比较准
            json2_data = json1_to_json2(data)
            json2_str = json.dumps(json2_data, ensure_ascii=False)
            data_lines.append(('\t'.join([img_file.relative_to(data_dir).as_posix(), json2_str])))
        return data_lines



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 11:16

import base64
import concurrent.futures
import json
import os
import re
import subprocess
import time

from tqdm import tqdm

from pyxllib.text.newbie import add_quote


def mtqdm(func, iterable, *args, max_workers=1, check_per_seconds=0.01, **kwargs):
    """ 对tqdm的封装，增加了多线程的支持

    这里名称前缀多出的m有multi的意思

    :param max_workers: 默认是单线程，改成None会自动变为多线程
        或者可以自己指定线程数
        注意，使用负数，可以用对等绝对值数据的“多进程”
    :param smoothing: tqdm官方默认值是0.3
        这里关掉指数移动平均，直接计算整体平均速度
        因为对我个人来说，大部分时候需要严谨地分析性能，得到整体平均速度，而不是预估当前速度
    :param mininterval: 官方默认值是0.1，表示显示更新间隔秒数
        这里不用那么频繁，每秒更新就行了~~
    :param check_per_seconds: 每隔多少秒检查队列
        有些任务，这个值故意设大一点，可以减少频繁的队列检查时间，提高运行速度
    整体功能类似Iterate
    """

    # 0 个人习惯参数
    kwargs['smoothing'] = kwargs.get('smoothing', 0)
    kwargs['mininterval'] = kwargs.get('mininterval', 1)

    if max_workers == 1:
        # 1 如果只用一个线程，则不使用concurrent.futures.ThreadPoolExecutor，能加速
        for x in tqdm(iterable, *args, **kwargs):
            func(x)
    else:
        # 2 默认的多线程运行机制，出错是不会暂停的；这里对原函数功能进行封装，增加报错功能
        error = False

        def wrap_func(x):
            nonlocal error
            try:
                func(x)
            except Exception as e:
                error = e

        # 3 多线程/多进程 和 进度条 功能的结合
        if max_workers > 1:
            executor = concurrent.futures.ThreadPoolExecutor(max_workers)
            for x in tqdm(iterable, *args, **kwargs):
                while executor._work_queue.qsize():
                    if check_per_seconds:
                        time.sleep(check_per_seconds)
                executor.submit(wrap_func, x)
                if error:
                    raise error
        else:
            executor = concurrent.futures.ProcessPoolExecutor(-max_workers)
            for x in tqdm(iterable, *args, **kwargs):
                # while executor._call_queue.pending_work_items:
                #     if check_per_seconds:
                #         time.sleep(check_per_seconds)
                executor.submit(wrap_func, x)
                if error:
                    raise error

        executor.shutdown()


def distribute_package(root, version=None, repository=None, *, upload=True):
    """ 发布包的工具函数

    :param root: 项目的根目录，例如 'D:/slns/pyxllib'
        根目录下有对应的 setup.py 等文件
    :param repository: 比如我配置了 [xlpr]，就可以传入 'xlpr'
    """
    from pyxllib.file.specialist import File, Dir

    # 1 切换工作目录
    os.chdir(str(root))

    # 2 改版本号
    if version:
        f = File('setup.py', root)
        s = re.sub(r"(version\s*=\s*)(['\"])(.+?)(\2)", fr'\1\g<2>{version}\4', f.read())
        f.write(s)

    # 3 打包
    subprocess.run('python setup.py sdist')

    # 4 上传
    if upload:
        # 上传
        cmd = 'twine upload dist/*'
        if repository:
            cmd += f' -r {repository}'
        subprocess.run(cmd)
        # 删除打包生成的中间文件
        Dir('dist').delete()
        Dir('build').delete()

        # 这个不能删，不然importlib会读取不到模块的版本号
        # [d.delete() for d in Dir('.').select_dirs(r'*.egg-info')]


class XlOsEnv:
    """ pyxllib库自带的一套环境变量数据解析类

    会将json的字符串值，或者普通str，存储到环境变量中

    环境变量也可以用来实现全局变量的信息传递，虽然不太建议这样做

    >> XlOsEnv.persist_set('TP10_ACCOUNT',
                           {'server': '172.16.250.250', 'port': 22, 'user': 'ckz', 'passwd': '123456'},
                           True)
    >> print(XlOsEnv.get('TP10_ACCOUNT'), True)  # 展示存储的账号信息
    eyJzZXJ2ZXIiOiAiMTcyLjE2LjE3MC4xMzQiLCAicG9ydCI6IDIyLCAidXNlciI6ICJjaGVua3VuemUiLCAicGFzc3dkIjogImNvZGV4bHByIn0=
    >> XlOsEnv.unset('TP10_ACCOUNT')
    """

    @classmethod
    def get(cls, name, *, decoding=False):
        """ 获取环境变量值

        :param name: 环境变量名
        :param decoding: 是否需要先进行base64解码
        :return:
            返回json解析后的数据
            或者普通的字符串值
        """
        value = os.getenv(name, None)
        if value is None:
            return value

        if decoding:
            value = base64.b64decode(value.encode())

        try:
            return json.loads(value)
        except json.decoder.JSONDecodeError:
            return value

    @classmethod
    def set(cls, name, value, encoding=False):
        """ 临时改变环境变量

        :param name: 环境变量名
        :param value: 要存储的值
        :param encoding: 是否将内容转成base64后，再存储环境变量
            防止一些密码信息，明文写出来太容易泄露
            不过这个策略也很容易被破解；只防君子，难防小人

            当然，谁看到这有闲情功夫的话，可以考虑做一套更复杂的加密系统
            并且encoding支持多种不同的解加密策略，这样单看环境变量值就很难破译了
        :return: str, 最终存储的字符串内容
        """
        # 1 打包
        if isinstance(value, str):
            value = add_quote(value)
        else:
            value = json.dumps(value)

        # 2 编码
        if encoding:
            value = base64.b64encode(value.encode()).decode()

        # 3 存储到环境变量
        os.environ[name] = value

        return value

    @classmethod
    def persist_set(cls, name, value, encoding=False):
        """ python里默认是改不了系统变量的，需要使用一些特殊手段

        https://stackoverflow.com/questions/17657686/is-it-possible-to-set-an-environment-variable-from-python-permanently/17657905
        """
        # 写入环境变量这里是有点小麻烦的，要考虑unix和windows不同平台，以及怎么持久化存储的问题，这里直接调用一个三方库来解决
        from envariable import setenv

        value = cls.set(name, value, encoding)
        if value[0] == value[-1] == '"':
            value = '\\' + value + '\\'
        setenv(name, value)

    @classmethod
    def unset(cls, name):
        """ 删除 """
        from envariable import unsetenv
        unsetenv(name)



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from PIL import Image
import numpy as np
from ppocr.tools.infer.utility import draw_ocr_box_txt, init_args as infer_args


def init_args():
    parser = infer_args()

    # params for output
    parser.add_argument("--output", type=str, default='./output')
    # params for table structure
    parser.add_argument("--table_max_len", type=int, default=488)
    parser.add_argument("--table_model_dir", type=str)
    parser.add_argument("--table_char_type", type=str, default='en')
    parser.add_argument(
        "--table_char_dict_path",
        type=str,
        default="../ppocr/utils/dict/table_structure_dict.txt")
    parser.add_argument(
        "--layout_path_model",
        type=str,
        default="lp://PubLayNet/ppyolov2_r50vd_dcn_365e_publaynet/config")

    # params for ser
    parser.add_argument("--model_name_or_path", type=str)
    parser.add_argument("--max_seq_length", type=int, default=512)
    parser.add_argument(
        "--label_map_path", type=str, default='./vqa/labels/labels_ser.txt')

    parser.add_argument(
        "--mode",
        type=str,
        default='structure',
        help='structure and vqa is supported')
    return parser


def parse_args():
    parser = init_args()
    return parser.parse_args()


def draw_structure_result(image, result, font_path):
    if isinstance(image, np.ndarray):
        image = Image.fromarray(image)
    boxes, txts, scores = [], [], []
    for region in result:
        if region['type'] == 'Table':
            pass
        else:
            for box, rec_res in zip(region['res'][0], region['res'][1]):
                boxes.append(np.array(box).reshape(-1, 2))
                txts.append(rec_res[0])
                scores.append(rec_res[1])
    im_show = draw_ocr_box_txt(
        image, boxes, txts, scores, font_path=font_path, drop_score=0)
    return im_show



# copyright (c) 2020 PaddlePaddle Authors. All Rights Reserve.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import sys
import tarfile
import requests
from tqdm import tqdm

from ppocr.utils.logging import get_logger


def download_with_progressbar(url, save_path):
    logger = get_logger()
    response = requests.get(url, stream=True)
    if response.status_code == 200:
        total_size_in_bytes = int(response.headers.get('content-length', 1))
        block_size = 1024  # 1 Kibibyte
        progress_bar = tqdm(
            total=total_size_in_bytes, unit='iB', unit_scale=True)
        with open(save_path, 'wb') as file:
            for data in response.iter_content(block_size):
                progress_bar.update(len(data))
                file.write(data)
        progress_bar.close()
    else:
        logger.error("Something went wrong while downloading models")
        sys.exit(0)


def maybe_download(model_storage_directory, url):
    # using custom model
    tar_file_name_list = [
        'inference.pdiparams', 'inference.pdiparams.info', 'inference.pdmodel'
    ]
    if not os.path.exists(
            os.path.join(model_storage_directory, 'inference.pdiparams')
    ) or not os.path.exists(
            os.path.join(model_storage_directory, 'inference.pdmodel')):
        assert url.endswith('.tar'), 'Only supports tar compressed package'
        tmp_path = os.path.join(model_storage_directory, url.split('/')[-1])
        print('download {} to {}'.format(url, tmp_path))
        os.makedirs(model_storage_directory, exist_ok=True)
        download_with_progressbar(url, tmp_path)
        with tarfile.open(tmp_path, 'r') as tarObj:
            for member in tarObj.getmembers():
                filename = None
                for tar_file_name in tar_file_name_list:
                    if tar_file_name in member.name:
                        filename = tar_file_name
                if filename is None:
                    continue
                file = tarObj.extractfile(member)
                with open(
                        os.path.join(model_storage_directory, filename),
                        'wb') as f:
                    f.write(file.read())
        os.remove(tmp_path)


def is_link(s):
    return s is not None and s.startswith('http')


def confirm_model_dir_url(model_dir, default_model_dir, default_url):
    url = default_url
    if model_dir is None or is_link(model_dir):
        if is_link(model_dir):
            url = model_dir
        file_name = url.split('/')[-1][:-4]
        model_dir = default_model_dir
        model_dir = os.path.join(model_dir, file_name)
    return model_dir, url



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys
import pickle

__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '..')))

from ppocr.data import build_dataloader
from ppocr.modeling.architectures import build_model
from ppocr.postprocess import build_post_process
from ppocr.utils.save_load import load_model
from ppocr.utils.utility import print_dict
import tools.program as program


def main():
    global_config = config['Global']
    # build dataloader
    config['Eval']['dataset']['name'] = config['Train']['dataset']['name']
    config['Eval']['dataset']['data_dir'] = config['Train']['dataset'][
        'data_dir']
    config['Eval']['dataset']['label_file_list'] = config['Train']['dataset'][
        'label_file_list']
    eval_dataloader = build_dataloader(config, 'Eval', device, logger)

    # build post process
    post_process_class = build_post_process(config['PostProcess'],
                                            global_config)

    # build model
    # for rec algorithm
    if hasattr(post_process_class, 'character'):
        char_num = len(getattr(post_process_class, 'character'))
        config['Architecture']["Head"]['out_channels'] = char_num

    #set return_features = True
    config['Architecture']["Head"]["return_feats"] = True

    model = build_model(config['Architecture'])

    best_model_dict = load_model(config, model)
    if len(best_model_dict):
        logger.info('metric in ckpt ***************')
        for k, v in best_model_dict.items():
            logger.info('{}:{}'.format(k, v))

    # get features from train data
    char_center = program.get_center(model, eval_dataloader, post_process_class)

    #serialize to disk
    with open("train_center.pkl", 'wb') as f:
        pickle.dump(char_center, f)
    return


if __name__ == '__main__':
    config, device, logger, vdl_writer = program.preprocess()
    main()



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

import os
import sys

__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(__dir__)
sys.path.append(os.path.abspath(os.path.join(__dir__, '..')))

os.environ["FLAGS_allocator_strategy"] = 'auto_growth'

import paddle

from ppocr.data import create_operators, transform
from ppocr.modeling.architectures import build_model
from ppocr.postprocess import build_post_process
from ppocr.utils.save_load import load_model
from ppocr.utils.utility import get_image_file_list
import tools.program as program


def main():
    global_config = config['Global']

    # build post process
    post_process_class = build_post_process(config['PostProcess'],
                                            global_config)

    # build model
    model = build_model(config['Architecture'])

    load_model(config, model)

    # create data ops
    transforms = []
    for op in config['Eval']['dataset']['transforms']:
        op_name = list(op)[0]
        if 'Label' in op_name:
            continue
        elif op_name == 'KeepKeys':
            op[op_name]['keep_keys'] = ['image']
        transforms.append(op)
    global_config['infer_mode'] = True
    ops = create_operators(transforms, global_config)

    model.eval()
    for file in get_image_file_list(config['Global']['infer_img']):
        logger.info("infer_img: {}".format(file))
        with open(file, 'rb') as f:
            img = f.read()
            data = {'image': img}
        batch = transform(data, ops)

        images = np.expand_dims(batch[0], axis=0)
        images = paddle.to_tensor(images)
        preds = model(images)
        post_result = post_process_class(preds)
        for rec_reuslt in post_result:
            logger.info('\t result: {}'.format(rec_reuslt))
    logger.info("success!")


if __name__ == '__main__':
    config, device, logger, vdl_writer = program.preprocess()
    main()



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import sys

__dir__ = os.path.dirname(__file__)
sys.path.append(__dir__)
sys.path.append(os.path.join(__dir__, '..'))

import numpy as np
from .locality_aware_nms import nms_locality
import paddle
import cv2
import time


class SASTPostProcess(object):
    """
    The post process for SAST.
    """

    def __init__(self,
                 score_thresh=0.5,
                 nms_thresh=0.2,
                 sample_pts_num=2,
                 shrink_ratio_of_width=0.3,
                 expand_scale=1.0,
                 tcl_map_thresh=0.5,
                 **kwargs):

        self.score_thresh = score_thresh
        self.nms_thresh = nms_thresh
        self.sample_pts_num = sample_pts_num
        self.shrink_ratio_of_width = shrink_ratio_of_width
        self.expand_scale = expand_scale
        self.tcl_map_thresh = tcl_map_thresh

        # c++ la-nms is faster, but only support python 3.5
        self.is_python35 = False
        if sys.version_info.major == 3 and sys.version_info.minor == 5:
            self.is_python35 = True

    def point_pair2poly(self, point_pair_list):
        """
        Transfer vertical point_pairs into poly point in clockwise.
        """
        # constract poly
        point_num = len(point_pair_list) * 2
        point_list = [0] * point_num
        for idx, point_pair in enumerate(point_pair_list):
            point_list[idx] = point_pair[0]
            point_list[point_num - 1 - idx] = point_pair[1]
        return np.array(point_list).reshape(-1, 2)

    def shrink_quad_along_width(self,
                                quad,
                                begin_width_ratio=0.,
                                end_width_ratio=1.):
        """ 
        Generate shrink_quad_along_width.
        """
        ratio_pair = np.array(
            [[begin_width_ratio], [end_width_ratio]], dtype=np.float32)
        p0_1 = quad[0] + (quad[1] - quad[0]) * ratio_pair
        p3_2 = quad[3] + (quad[2] - quad[3]) * ratio_pair
        return np.array([p0_1[0], p0_1[1], p3_2[1], p3_2[0]])

    def expand_poly_along_width(self, poly, shrink_ratio_of_width=0.3):
        """
        expand poly along width.
        """
        point_num = poly.shape[0]
        left_quad = np.array(
            [poly[0], poly[1], poly[-2], poly[-1]], dtype=np.float32)
        left_ratio = -shrink_ratio_of_width * np.linalg.norm(left_quad[0] - left_quad[3]) / \
                     (np.linalg.norm(left_quad[0] - left_quad[1]) + 1e-6)
        left_quad_expand = self.shrink_quad_along_width(left_quad, left_ratio,
                                                        1.0)
        right_quad = np.array(
            [
                poly[point_num // 2 - 2], poly[point_num // 2 - 1],
                poly[point_num // 2], poly[point_num // 2 + 1]
            ],
            dtype=np.float32)
        right_ratio = 1.0 + \
                      shrink_ratio_of_width * np.linalg.norm(right_quad[0] - right_quad[3]) / \
                      (np.linalg.norm(right_quad[0] - right_quad[1]) + 1e-6)
        right_quad_expand = self.shrink_quad_along_width(right_quad, 0.0,
                                                         right_ratio)
        poly[0] = left_quad_expand[0]
        poly[-1] = left_quad_expand[-1]
        poly[point_num // 2 - 1] = right_quad_expand[1]
        poly[point_num // 2] = right_quad_expand[2]
        return poly

    def restore_quad(self, tcl_map, tcl_map_thresh, tvo_map):
        """Restore quad."""
        xy_text = np.argwhere(tcl_map[:, :, 0] > tcl_map_thresh)
        xy_text = xy_text[:, ::-1]  # (n, 2)

        # Sort the text boxes via the y axis
        xy_text = xy_text[np.argsort(xy_text[:, 1])]

        scores = tcl_map[xy_text[:, 1], xy_text[:, 0], 0]
        scores = scores[:, np.newaxis]

        # Restore
        point_num = int(tvo_map.shape[-1] / 2)
        assert point_num == 4
        tvo_map = tvo_map[xy_text[:, 1], xy_text[:, 0], :]
        xy_text_tile = np.tile(xy_text, (1, point_num))  # (n, point_num * 2)
        quads = xy_text_tile - tvo_map

        return scores, quads, xy_text

    def quad_area(self, quad):
        """
        compute area of a quad.
        """
        edge = [(quad[1][0] - quad[0][0]) * (quad[1][1] + quad[0][1]),
                (quad[2][0] - quad[1][0]) * (quad[2][1] + quad[1][1]),
                (quad[3][0] - quad[2][0]) * (quad[3][1] + quad[2][1]),
                (quad[0][0] - quad[3][0]) * (quad[0][1] + quad[3][1])]
        return np.sum(edge) / 2.

    def nms(self, dets):
        if self.is_python35:
            import lanms
            dets = lanms.merge_quadrangle_n9(dets, self.nms_thresh)
        else:
            dets = nms_locality(dets, self.nms_thresh)
        return dets

    def cluster_by_quads_tco(self, tcl_map, tcl_map_thresh, quads, tco_map):
        """
        Cluster pixels in tcl_map based on quads.
        """
        instance_count = quads.shape[0] + 1  # contain background
        instance_label_map = np.zeros(tcl_map.shape[:2], dtype=np.int32)
        if instance_count == 1:
            return instance_count, instance_label_map

        # predict text center
        xy_text = np.argwhere(tcl_map[:, :, 0] > tcl_map_thresh)
        n = xy_text.shape[0]
        xy_text = xy_text[:, ::-1]  # (n, 2)
        tco = tco_map[xy_text[:, 1], xy_text[:, 0], :]  # (n, 2)
        pred_tc = xy_text - tco

        # get gt text center
        m = quads.shape[0]
        gt_tc = np.mean(quads, axis=1)  # (m, 2)

        pred_tc_tile = np.tile(pred_tc[:, np.newaxis, :],
                               (1, m, 1))  # (n, m, 2)
        gt_tc_tile = np.tile(gt_tc[np.newaxis, :, :], (n, 1, 1))  # (n, m, 2)
        dist_mat = np.linalg.norm(pred_tc_tile - gt_tc_tile, axis=2)  # (n, m)
        xy_text_assign = np.argmin(dist_mat, axis=1) + 1  # (n,)

        instance_label_map[xy_text[:, 1], xy_text[:, 0]] = xy_text_assign
        return instance_count, instance_label_map

    def estimate_sample_pts_num(self, quad, xy_text):
        """
        Estimate sample points number.
        """
        eh = (np.linalg.norm(quad[0] - quad[3]) +
              np.linalg.norm(quad[1] - quad[2])) / 2.0
        ew = (np.linalg.norm(quad[0] - quad[1]) +
              np.linalg.norm(quad[2] - quad[3])) / 2.0

        dense_sample_pts_num = max(2, int(ew))
        dense_xy_center_line = xy_text[np.linspace(
            0,
            xy_text.shape[0] - 1,
            dense_sample_pts_num,
            endpoint=True,
            dtype=np.float32).astype(np.int32)]

        dense_xy_center_line_diff = dense_xy_center_line[
            1:] - dense_xy_center_line[:-1]
        estimate_arc_len = np.sum(
            np.linalg.norm(
                dense_xy_center_line_diff, axis=1))

        sample_pts_num = max(2, int(estimate_arc_len / eh))
        return sample_pts_num

    def detect_sast(self,
                    tcl_map,
                    tvo_map,
                    tbo_map,
                    tco_map,
                    ratio_w,
                    ratio_h,
                    src_w,
                    src_h,
                    shrink_ratio_of_width=0.3,
                    tcl_map_thresh=0.5,
                    offset_expand=1.0,
                    out_strid=4.0):
        """
        first resize the tcl_map, tvo_map and tbo_map to the input_size, then restore the polys
        """
        # restore quad
        scores, quads, xy_text = self.restore_quad(tcl_map, tcl_map_thresh,
                                                   tvo_map)
        dets = np.hstack((quads, scores)).astype(np.float32, copy=False)
        dets = self.nms(dets)
        if dets.shape[0] == 0:
            return []
        quads = dets[:, :-1].reshape(-1, 4, 2)

        # Compute quad area
        quad_areas = []
        for quad in quads:
            quad_areas.append(-self.quad_area(quad))

        # instance segmentation
        # instance_count, instance_label_map = cv2.connectedComponents(tcl_map.astype(np.uint8), connectivity=8)
        instance_count, instance_label_map = self.cluster_by_quads_tco(
            tcl_map, tcl_map_thresh, quads, tco_map)

        # restore single poly with tcl instance.
        poly_list = []
        for instance_idx in range(1, instance_count):
            xy_text = np.argwhere(instance_label_map == instance_idx)[:, ::-1]
            quad = quads[instance_idx - 1]
            q_area = quad_areas[instance_idx - 1]
            if q_area < 5:
                continue

            #
            len1 = float(np.linalg.norm(quad[0] - quad[1]))
            len2 = float(np.linalg.norm(quad[1] - quad[2]))
            min_len = min(len1, len2)
            if min_len < 3:
                continue

            # filter small CC
            if xy_text.shape[0] <= 0:
                continue

            # filter low confidence instance
            xy_text_scores = tcl_map[xy_text[:, 1], xy_text[:, 0], 0]
            if np.sum(xy_text_scores) / quad_areas[instance_idx - 1] < 0.1:
                # if np.sum(xy_text_scores) / quad_areas[instance_idx - 1] < 0.05:
                continue

            # sort xy_text
            left_center_pt = np.array(
                [[(quad[0, 0] + quad[-1, 0]) / 2.0,
                  (quad[0, 1] + quad[-1, 1]) / 2.0]])  # (1, 2)
            right_center_pt = np.array(
                [[(quad[1, 0] + quad[2, 0]) / 2.0,
                  (quad[1, 1] + quad[2, 1]) / 2.0]])  # (1, 2)
            proj_unit_vec = (right_center_pt - left_center_pt) / \
                            (np.linalg.norm(right_center_pt - left_center_pt) + 1e-6)
            proj_value = np.sum(xy_text * proj_unit_vec, axis=1)
            xy_text = xy_text[np.argsort(proj_value)]

            # Sample pts in tcl map
            if self.sample_pts_num == 0:
                sample_pts_num = self.estimate_sample_pts_num(quad, xy_text)
            else:
                sample_pts_num = self.sample_pts_num
            xy_center_line = xy_text[np.linspace(
                0,
                xy_text.shape[0] - 1,
                sample_pts_num,
                endpoint=True,
                dtype=np.float32).astype(np.int32)]

            point_pair_list = []
            for x, y in xy_center_line:
                # get corresponding offset
                offset = tbo_map[y, x, :].reshape(2, 2)
                if offset_expand != 1.0:
                    offset_length = np.linalg.norm(
                        offset, axis=1, keepdims=True)
                    expand_length = np.clip(
                        offset_length * (offset_expand - 1),
                        a_min=0.5,
                        a_max=3.0)
                    offset_detal = offset / offset_length * expand_length
                    offset = offset + offset_detal
                    # original point
                ori_yx = np.array([y, x], dtype=np.float32)
                point_pair = (ori_yx + offset)[:, ::-1] * out_strid / np.array(
                    [ratio_w, ratio_h]).reshape(-1, 2)
                point_pair_list.append(point_pair)

            # ndarry: (x, 2), expand poly along width
            detected_poly = self.point_pair2poly(point_pair_list)
            detected_poly = self.expand_poly_along_width(detected_poly,
                                                         shrink_ratio_of_width)
            detected_poly[:, 0] = np.clip(
                detected_poly[:, 0], a_min=0, a_max=src_w)
            detected_poly[:, 1] = np.clip(
                detected_poly[:, 1], a_min=0, a_max=src_h)
            poly_list.append(detected_poly)

        return poly_list

    def __call__(self, outs_dict, shape_list):
        score_list = outs_dict['f_score']
        border_list = outs_dict['f_border']
        tvo_list = outs_dict['f_tvo']
        tco_list = outs_dict['f_tco']
        if isinstance(score_list, paddle.Tensor):
            score_list = score_list.numpy()
            border_list = border_list.numpy()
            tvo_list = tvo_list.numpy()
            tco_list = tco_list.numpy()

        img_num = len(shape_list)
        poly_lists = []
        for ino in range(img_num):
            p_score = score_list[ino].transpose((1, 2, 0))
            p_border = border_list[ino].transpose((1, 2, 0))
            p_tvo = tvo_list[ino].transpose((1, 2, 0))
            p_tco = tco_list[ino].transpose((1, 2, 0))
            src_h, src_w, ratio_h, ratio_w = shape_list[ino]

            poly_list = self.detect_sast(
                p_score,
                p_tvo,
                p_border,
                p_tco,
                ratio_w,
                ratio_h,
                src_w,
                src_h,
                shrink_ratio_of_width=self.shrink_ratio_of_width,
                tcl_map_thresh=self.tcl_map_thresh,
                offset_expand=self.expand_scale)
            poly_lists.append({'points': np.array(poly_list)})

        return poly_lists



# Copyright (c) 2020 PaddlePaddle Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import numpy as np
import paddle
import signal
import random

__dir__ = os.path.dirname(os.path.abspath(__file__))
sys.path.append(os.path.abspath(os.path.join(__dir__, '../..')))

import copy
from paddle.io import Dataset, DataLoader, BatchSampler, DistributedBatchSampler
import paddle.distributed as dist

from ppocr.data.imaug import transform, create_operators
# 可以在ppocr/data目录下新增脚本，添加自己的数据格式
# 个人想法，到时候新增的数据格式类，都统一放到一个文件xl_dataset里，方便整理和分享
from ppocr.data.simple_dataset import *  # 这里扩展了一些自己的基础数据格式
from ppocr.data.lmdb_dataset import LMDBDataSet
from ppocr.data.pgnet_dataset import PGDataSet
from ppocr.data.pubtab_dataset import PubTabDataSet

# 或者不在ppocr.data这里加也可以，重点是能导入特定接口范式的类，让这里eval能取到即可。

__all__ = ['build_dataloader', 'transform', 'create_operators']


def term_mp(sig_num, frame):
    """ kill all child processes
    """
    pid = os.getpid()
    pgid = os.getpgid(os.getpid())
    print("main proc {} exit, kill process group " "{}".format(pid, pgid))
    os.killpg(pgid, signal.SIGKILL)


def build_dataloader(config, mode, device, logger, seed=None):
    config = copy.deepcopy(config)

    support_dict = [
        'SimpleDataSet', 'LMDBDataSet', 'PGDataSet', 'PubTabDataSet', 'XlSimpleDataSet'
    ]
    module_name = config[mode]['dataset']['name']
    assert module_name in support_dict, Exception(
        'DataSet only support {}'.format(support_dict))
    # 这里eval没有安全隐患，因为会提前判断module_name要属于support_dict中的值
    assert mode in ['Train', 'Eval', 'Test'
                    ], "Mode should be Train, Eval or Test."

    # 自定义数据格式类，除了初始化，似乎没看到还有其他框架性的约束
    dataset = eval(module_name)(config, mode, logger, seed)
    loader_config = config[mode]['loader']
    # loader的必填参数
    batch_size = loader_config['batch_size_per_card']
    drop_last = loader_config['drop_last']
    shuffle = loader_config['shuffle']
    num_workers = loader_config['num_workers']
    # 可选参数
    # 注意这里实现机制策略与d2的不同，d2习惯是把这个参数加入到初始默认配置字典中，而ppocr是在函数运算中智能判断。
    if 'use_shared_memory' in loader_config.keys():
        use_shared_memory = loader_config['use_shared_memory']
    else:
        # use_shared_memory (bool) - 是否使用共享内存来提升子进程将数据放入进程间队列的速度，
        # 该参数尽在多进程模式下有效(即 num_workers > 0 )，
        # 请确认机器上有足够的共享内存空间(如Linux系统下 /dev/shm/ 目录空间大小)再设置此参数。
        # 默认为True。
        use_shared_memory = True
    if mode == "Train":
        # Train会用多卡机制分配BatchSampler，当然，如果外部设置了单卡也可以，单卡是特殊的多卡机制。
        # Distribute data to multiple cards
        batch_sampler = DistributedBatchSampler(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            drop_last=drop_last)
    else:
        # 非Train阶段，强制使用单卡处理。
        # Distribute data to single card
        batch_sampler = BatchSampler(
            dataset=dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            drop_last=drop_last)

    data_loader = DataLoader(
        dataset=dataset,
        batch_sampler=batch_sampler,
        places=device,
        num_workers=num_workers,
        return_list=True,
        use_shared_memory=use_shared_memory)

    # 看到用了signal库，好奇~~ https://www.jianshu.com/p/e0a69beb98bb
    # support exit using ctrl+c
    signal.signal(signal.SIGINT, term_mp)
    signal.signal(signal.SIGTERM, term_mp)

    return data_loader


def build_dataset(config, mode, logger, seed=None):
    """ ckz: 有时候不需要获得loader，只要dataset即可 """
    config = copy.deepcopy(config)

    support_dict = [
        'SimpleDataSet', 'LMDBDataSet', 'PGDataSet', 'PubTabDataSet', 'XlSimpleDataSet'
    ]
    module_name = config[mode]['dataset']['name']
    assert module_name in support_dict, Exception(
        'DataSet only support {}'.format(support_dict))
    assert mode in ['Train', 'Eval', 'Test'
                    ], "Mode should be Train, Eval or Test."

    dataset = eval(module_name)(config, mode, logger, seed)

    return dataset



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Author : 陈坤泽
# @Email  : 877362867@qq.com
# @Date   : 2021/06/06 23:10


from pyxllib.xlcv import *

import torch
from torch import nn, optim
import torch.utils.data

import torchvision
from torchvision import transforms
from easydict import EasyDict

# 把pytorch等常用的导入写了
import torch.utils.data
from torchvision.datasets import VisionDataset

from pyxllib.ai.specialist import ClasEvaluater, NvmDevice

__base = """
"""


def get_most_free_torch_gpu_device():
    gpu_id = NvmDevice().get_most_free_gpu_id()
    if gpu_id is not None:
        return torch.device(f'cuda:{gpu_id}')


def get_device():
    """ 自动获得一个可用的设备
    """
    return get_most_free_torch_gpu_device() or torch.device('cpu')


__data = """
"""


class TinyDataset(torch.utils.data.Dataset):
    def __init__(self, labelfile, label_transform, maxn=None):
        """ 超轻量级的Dataset类，一般由外部ProjectData类指定每行label的转换规则 """
        self.labels = File(labelfile).read().splitlines()
        self.label_transform = label_transform

        self.number = len(self.labels)
        if maxn: self.number = min(self.number, maxn)

    def __len__(self):
        return self.number

    def __getitem__(self, idx):
        return self.label_transform(self.labels[idx])


class InputDataset(torch.utils.data.Dataset):
    def __init__(self, raw_in, transform=None, *, y_placeholder=...):
        """ 将非list、tuple数据转为list，并生成一个dataset类的万用类
        :param raw_in:
        """
        if not isinstance(raw_in, (list, tuple)):
            raw_in = [raw_in]

        self.data = raw_in
        self.transform = transform
        self.y_placeholder = y_placeholder

    def __len__(self):
        return len(self.data)

    def __getitem__(self, idx):
        x = self.data[idx]
        if self.transform:
            x = self.transform(x)

        if self.y_placeholder is not ...:
            return x, self.y_placeholder
        else:
            return x


__model = """
"""


class LeNet5(nn.Module):
    """ https://towardsdatascience.com/implementing-yann-lecuns-lenet-5-in-pytorch-5e05a0911320 """

    def __init__(self, n_classes):
        super().__init__()

        self.feature_extractor = nn.Sequential(
            nn.Conv2d(1, 6, kernel_size=5, stride=1),
            nn.Tanh(),
            nn.AvgPool2d(kernel_size=2),
            nn.Conv2d(6, 16, kernel_size=5, stride=1),
            nn.Tanh(),
            nn.AvgPool2d(kernel_size=2),
            nn.Conv2d(16, 120, kernel_size=5, stride=1),
            nn.Tanh()
        )

        self.classifier = nn.Sequential(
            nn.Linear(120, 84),
            nn.Tanh(),
            nn.Linear(84, n_classes),
        )

    def forward(self, batched_inputs):
        device = next(self.parameters()).device

        x = batched_inputs[0].to(device)
        x = self.feature_extractor(x)
        x = torch.flatten(x, 1)
        logits = self.classifier(x)

        if self.training:
            y = batched_inputs[1].to(device)
            return nn.functional.cross_entropy(logits, y)
        else:
            return logits.argmax(dim=1)


__train = """
"""


class Trainer:
    """ 对pytorch模型的训练、测试等操作的进一步封装

    # TODO log变成可选项，可以关掉
    """

    def __init__(self, log_dir, device, data, model, optimizer,
                 loss_func=None, pred_func=None, accuracy_func=None):
        # 0 初始化成员变量
        self.log_dir, self.device = log_dir, device
        self.data, self.model, self.optimizer = data, model, optimizer
        if loss_func: self.loss_func = loss_func
        if pred_func: self.pred_func = pred_func
        if accuracy_func: self.accuracy_func = accuracy_func

        # 1 日志
        timetag = datetime.datetime.now().strftime('%Y%m%d.%H%M%S')
        # self.curlog_dir = Dir(self.log_dir / timetag)  # 本轮运行，实际log位置，是存放在一个子目录里
        self.curlog_dir = Dir(self.log_dir)
        self.curlog_dir.ensure_dir()
        self.log = get_xllog(log_file=self.curlog_dir / 'log.txt')
        self.log.info(f'1/4 log_dir={self.curlog_dir}')

        # 2 设备
        self.log.info(f'2/4 use_device={self.device}')

        # 3 数据
        self.train_dataloader = self.data.get_train_dataloader()
        self.val_dataloader = self.data.get_val_dataloader()
        self.train_data_number = len(self.train_dataloader.dataset)
        self.val_data_number = len(self.val_dataloader.dataset)
        self.log.info(f'3/4 get data, train_data_number={self.train_data_number}(batch={len(self.train_dataloader)}), '
                      f'val_data_number={self.val_data_number}(batch={len(self.val_dataloader)}), '
                      f'batch_size={self.data.batch_size}')

        # 4 模型
        parasize = sum(map(lambda p: p.numel(), self.model.parameters()))
        self.log.info(
            f'4/4 model parameters size: {parasize}* 4 Bytes per float ≈ {humanfriendly.format_size(parasize * 4)}')

        # 5 其他辅助变量
        self.min_total_loss = math.inf  # 目前epoch中总损失最小的值（训练损失，训练过程损失）
        self.min_train_loss = math.inf  # 训练集损失
        self.max_val_accuracy = 0  # 验证集精度

    @classmethod
    def loss_func(cls, model_out, y):
        """ 自定义损失函数 """
        # return loss
        raise NotImplementedError

    @classmethod
    def pred_func(cls, model_out):
        """ 自定义模型输出到预测结果 """
        # return y_hat
        raise NotImplementedError

    @classmethod
    def accuracy_func(cls, y_hat, y):
        """ 自定义预测结果和实际标签y之间的精度

        返回"正确的样本数"（在非分类任务中，需要抽象出这个数量关系）
        """
        # return accuracy
        raise NotImplementedError

    def loss_values_stat(self, loss_vales):
        """ 一组loss损失的统计分析

        :param loss_vales: 一次batch中，多份样本产生的误差数据
        :return: 统计信息文本字符串
        """
        if not loss_vales:
            raise ValueError

        data = np.array(loss_vales, dtype=float)
        n, sum_ = len(data), data.sum()
        mean, std = data.mean(), data.std()
        msg = f'total_loss={sum_:.3f}, mean±std={mean:.3f}±{std:.3f}({max(data):.3f}->{min(data):.3f})'
        if sum_ < self.min_total_loss:
            self.min_total_loss = sum_
            msg = '*' + msg
        return msg

    @classmethod
    def sample_size(cls, data):
        """ 单个样本占用的空间大小，返回字节数 """
        x, label = data.dataset[0]  # 取第0个样本作为参考
        return getasizeof(x.numpy()) + getasizeof(label)

    def save_model_state(self, file, if_exists='error'):
        """ 保存模型参数值
        一般存储model.state_dict，而不是直接存储model，确保灵活性

        # TODO 和path结合，增加if_exists参数
        """
        f = File(file, self.curlog_dir)
        if f.exist_preprcs(if_exists=if_exists):
            f.ensure_parent()
            torch.save(self.model.state_dict(), str(f))

    def load_model_state(self, file):
        """ 读取模型参数值

        注意load和save的root差异！ load的默认父目录是在log_dir，而save默认是在curlog_dir！
        """
        f = File(file, self.log_dir)
        self.model.load_state_dict(torch.load(str(f), map_location=self.device))

    def viz_data(self):
        """ 用visdom显示样本数据

        TODO 增加一些自定义格式参数
        TODO 不能使用\n、\r\n、<br/>实现文本换行，有时间可以研究下，结合nrow、图片宽度，自动推算，怎么美化展示效果
        """
        from visdom import Visdom

        viz = Visdom()
        if not viz: return

        x, label = next(iter(self.train_dataloader))
        viz.one_batch_images(x, label, 'train data')

        x, label = next(iter(self.val_dataloader))
        viz.one_batch_images(x, label, 'val data')

    def training_one_epoch(self):
        # 1 检查模式
        if not self.model.training:
            self.model.train(True)

        # 2 训练一轮
        loss_values = []
        for x, y in self.train_dataloader:
            # 每个batch可能很大，所以每个batch依次放到cuda，而不是一次性全放入
            x, y = x.to(self.device), y.to(self.device)

            y_hat = self.model(x)
            loss = self.loss_func(y_hat, y)
            loss_values.append(float(loss))

            self.optimizer.zero_grad()
            loss.backward()
            self.optimizer.step()

        # 3 训练阶段只看loss，不看实际预测准确度，默认每个epoch都会输出
        return loss_values

    def calculate_accuracy(self, dataloader):
        """ 测试验证集等数据上的精度 """
        # 1 eval模式
        if self.model.training:
            self.model.train(False)

        # 2 关闭梯度，可以节省显存和加速
        with torch.no_grad():
            tt = TicToc()

            # 预测结果，计算正确率
            loss, correct, number = [], 0, len(dataloader.dataset)
            for x, y in dataloader:
                x, y = x.to(self.device), y.to(self.device)
                model_out = self.model(x)
                loss.append(self.loss_func(model_out, y))
                y_hat = self.pred_func(model_out)
                correct += self.accuracy_func(y_hat, y)  # 预测正确的数量
            elapsed_time, mean_loss = tt.tocvalue(), np.mean(loss, dtype=float)
            accuracy = correct / number
            info = f'accuracy={correct:.0f}/{number} ({accuracy:.2%})\t' \
                   f'mean_loss={mean_loss:.3f}\telapsed_time={elapsed_time:.0f}s'
            return accuracy, mean_loss, info

    def train_accuracy(self):
        accuracy, mean_loss, info = self.calculate_accuracy(self.train_dataloader)
        info = 'train ' + info
        if mean_loss < self.min_train_loss:
            # 如果是best ever，则log换成debug模式输出
            self.log.debug('*' + info)
            self.min_train_loss = mean_loss
        else:
            self.log.info(info)
        return accuracy

    def val_accuracy(self, save_model=None):
        """
        :param save_model: 如果验证集精度best ever，则保存当前epoch模型
            如果精度不是最好的，哪怕指定save_model也不会保存的
        :return:
        """
        accuracy, mean_loss, info = self.calculate_accuracy(self.val_dataloader)
        info = '  val ' + info
        if accuracy > self.max_val_accuracy:
            self.log.debug('*' + info)
            if save_model:
                self.save_model_state(save_model, if_exists='replace')
            self.max_val_accuracy = accuracy
        else:
            self.log.info(info)
        return accuracy

    def training(self, epochs, *, start_epoch=0, log_interval=1):
        """ 主要训练接口

        :param epochs: 训练代数，输出时从1开始编号
        :param start_epoch: 直接从现有的第几个epoch的模型读取参数
            使用该参数，需要在self.save_dir有对应名称的model文件
        :param log_interval: 每隔几个epoch输出当前epoch的训练情况，损失值
            每个几轮epoch进行一次监控
            且如果总损失是训练以来最好的结果，则会保存模型
                并对训练集、测试集进行精度测试
        TODO 看到其他框架，包括智财的框架，对保存的模型文件，都有更规范的一套命名方案，有空要去学一下
        :return:
        """
        from visdom import Visdom

        # 1 配置参数
        tag = self.model.__class__.__name__
        epoch_time_tag = f'elapsed_time' if log_interval == 1 else f'{log_interval}*epoch_time'
        viz = Visdom()  # 其实这里不是用原生的Visdom，而是我封装过的，但是我封装的那个也没太大作用意义，删掉了

        # 2 加载之前的模型继续训练
        if start_epoch:
            self.load_model_state(f'{tag} epoch{start_epoch}.pth')

        # 3 训练
        tt = TicToc()
        for epoch in range(start_epoch + 1, epochs + 1):
            loss_values = self.training_one_epoch()
            # 3.1 训练损失可视化
            if viz: viz.loss_line(loss_values, epoch, 'train_loss')
            # 3.2 显示epoch训练效果
            if log_interval and epoch % log_interval == 0:
                # 3.2.1 显示训练用时、训练损失
                msg = self.loss_values_stat(loss_values)
                elapsed_time = tt.tocvalue(restart=True)
                info = f'epoch={epoch}, {epoch_time_tag}={elapsed_time:.0f}s\t{msg.lstrip("*")}'
                # 3.2.2 截止目前训练损失最小的结果
                if msg[0] == '*':
                    self.log.debug('*' + info)
                    # 3.2.2.1 测试训练集、验证集上的精度
                    accuracy1 = self.train_accuracy()
                    accuracy2 = self.val_accuracy(save_model=f'{tag} epoch{epoch}.pth')
                    # 3.2.2.2 可视化图表
                    if viz: viz.plot_line([[accuracy1, accuracy2]], [epoch], 'accuracy', legend=['train', 'val'])
                else:
                    self.log.info(info)


@deprecated(reason='推荐使用XlPredictor实现')
def gen_classification_func(model, *, state_file=None, transform=None, pred_func=None,
                            device=None):
    """ 工厂函数，生成一个分类器函数

    用这个函数做过渡的一个重要目的，也是避免重复加载模型

    :param model: 模型结构
    :param state_file: 存储参数的文件
    :param transform: 每一个输入样本的预处理函数
    :param pred_func: model 结果的参数的后处理
    :return: 返回的函数结构见下述 cls_func
    """
    if state_file: model.load_state_dict(torch.load(str(state_file), map_location=get_device()))
    model.train(False)
    device = device or get_device()
    model.to(device)

    def cls_func(raw_in):
        """
        :param raw_in: 输入可以是路径、np.ndarray、PIL图片等，都为转为batch结构的tensor
            im，一张图片路径、np.ndarray、PIL图片
            [im1, im2, ...]，多张图片清单
        :return: 输入如果只有一张图片，则返回一个结果
            否则会存在list，返回多个结果
        """
        dataset = InputDataset(raw_in, transform)
        # TODO batch_size根据device空间大小自适应设置
        xs = torch.utils.data.DataLoader(dataset, batch_size=8)
        res = None
        for x in xs:
            # 每个batch可能很大，所以每个batch依次放到cuda，而不是一次性全放入
            x = x.to(device)
            y = model(x)
            if pred_func: y = pred_func(y)
            res = y if res is None else (res + y)
        return res

    return cls_func


class XlPredictor:
    """ 生成一个类似函数用法的推断功能类

    这是一个通用的生成器，不同的业务可以继承开发，进一步设计细则

    这里默认写的结构是兼容detectron2框架的分类模型，即model.forward：
        输入：list，第1个是batch_x，第2个是batch_y
        输出：training是logits，eval是（batch）y_hat
    """

    def __init__(self, model, state_file=None, device=None, *, batch_size=1, y_placeholder=...):
        """
        :param model: 基于d2框架的模型结构
        :param state_file: 存储权重的文件
            一般写某个本地文件路径
            也可以写url地址，会下载存储到临时目录中
            可以不传入文件，直接给到初始化好权重的model，该模式常用语训练阶段的model
        :param batch_size: 支持每次最多几个样本一起推断
            具体运作细节参见 XlPredictor.inputs2loader的解释
            TODO batch_size根据device空间大小自适应设置
        :param y_placeholder: 参见XlPredictor.inputs2loader的解释
        """
        # 默认使用model所在的device
        if device is None:
            self.device = next(model.parameters()).device
        else:
            self.device = device

        if state_file is not None:
            if is_url(state_file):
                state_file = download(state_file, XlPath.tempdir() / 'xlpr')
            state = torch.load(str(state_file), map_location=self.device)
            if 'model' in state:
                state = state['model']
            model = model.to(device)
            model.load_state_dict(state)

        self.model = model
        self.model.train(False)

        self.batch_size = batch_size
        self.y_placeholder = y_placeholder

        self.transform = self.build_transform()
        self.target_transform = self.build_target_transform()

    @classmethod
    def build_transform(cls):
        """ 单个数据的转换规则，进入模型前的读取、格式转换

        为了效率性能，建议比较特殊的不同初始化策略，可以额外定义函数接口，例如：def from_paths()
        """
        return None

    @classmethod
    def build_target_transform(cls):
        """ 单个结果的转换的规则，模型预测完的结果，到最终结果的转换方式

        一些简单的情况直接返回y即可，但还有些复杂的任务可能要增加后处理
        """
        return None

    def inputs2loader(self, raw_in, *, batch_size=None, y_placeholder=..., sampler=None, **kwargs):
        """ 将各种类列表数据，转成torch.utils.data.DataLoader类型

        :param raw_in: 各种类列表数据格式，或者单个数据，都为转为batch结构的tensor
            torch.util.data.DataLoader
                此时XlPredictor自定义参数全部无效：transform、batch_size、y_placeholder，sampler
                因为这些在loader里都有配置了
            torch.util.data.Dataset
                此时可以定制扩展的参数有：batch_size，sampler
            [data1, data2, ...]，列表表示批量处理多个数据
                此时所有配置参数均可用：transform、batch_size、y_placeholder, sampler
                通常是图片文件路径清单
                    XlPredictor原生并没有扩展图片读取功能，但可以通过transform增加CvPrcs.read来支持
            single_data，单个数据
                通常是单个图片文件路径，注意transfrom要增加xlcv.read或xlpil.read来支持路径读取
                注意：有时候单个数据就是list格式，此时需要麻烦点，再套一层list避免歧义
        :param batch_size: 支持每次最多几个样本一起推断
            TODO batch_size根据device空间大小自适应设置
        :param y_placeholder: 常见的model.forward，是只输入batch_x就行，这时候就默认值处理机制就行
            但我从d2框架模仿的写法，forward需要补一个y的真实值，输入是[batch_x, batch_y]
            实际预测数据可能没有y，此时需要填充一个batch_y=None来对齐，即设置y_placeholder=None
            或者y_placeholder=0，则所有的y用0作为占位符填充
            不过用None、0、False这些填充都很诡异，容易误导开发者，建议需要设置的时候使用-1

            如果读者写的model.forward前传机制不同，本来batch_inputs就只输入x没有y，则这里不用设置y_placeholder参数
        :param sampler: 有时候只是要简单抽样部分数据测试下，可以设置该参数
            比如random.sample(range(10), 5)：可以从前10个数据中，无放回随机抽取5个数据
        """
        if isinstance(raw_in, torch.utils.data.DataLoader):
            loader = raw_in
        else:
            if not isinstance(raw_in, torch.utils.data.Dataset):
                y_placeholder = first_nonnone([y_placeholder, self.y_placeholder], lambda x: x is not ...)
                dataset = InputDataset(raw_in, self.transform, y_placeholder=y_placeholder)
            else:
                if not isinstance(raw_in, (list, tuple)):
                    raw_in = [raw_in]
                dataset = raw_in
            batch_size = first_nonnone([batch_size, self.batch_size])
            loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, sampler=sampler, **kwargs)

        return loader

    def forward(self, loader, *, info=False, return_gt=True):
        """ 前向传播

        改功能是__call__的子部分，常在train、eval阶段单独调用
        因为eval阶段，已经有预设好的train_loader、val_loader，不需要使用inputs2loader智能生成一个loader

        :param torch.utils.data.DataLoader loader: 标准的DataLoader类型，每次能获取[batch_x, batch_y]
        :param info: 有时候数据量比较大，可能会需要看推断进度条
        :param return_gt: 注意跟__call__的不同，这里默认是True，__call__默认是False
            前者常用于评价阶段，后者常用于部署阶段，应用场景不同，常见配置有区别
        :return:
            return_gt=True（默认）：[(y1, y_hat1), (y2, y_hat2), ...]
            return_gt=False：[y_hat1, y_hat2, ...]
        """
        preds = []
        with torch.no_grad():
            for batched_inputs in tqdm(loader, 'eval batch', disable=not info):
                # 有的模型forward里没有处理input的device问题，则需要在这里使用self.device设置
                # batched_inputs = batched_inputs.to(self.device)  # 这一步可能不应该写在这里，还是先注释掉吧
                batch_y = self.model(batched_inputs).tolist()
                if self.target_transform:
                    batch_y = [self.target_transform(y) for y in batch_y]
                if return_gt:
                    gt = batched_inputs[1].tolist()
                    preds += list(zip(*[gt, batch_y]))
                else:
                    preds += batch_y
        return preds

    def __call__(self, raw_in, *, batch_size=None, y_placeholder=...,
                 info=False, return_gt=False):
        """ 前传推断结果

        :param batch_size: 具体运行中可以重新指定batch_size
        :param return_gt: 使用该功能，必须确保每次loader都含有[x,y]，可能是raw_in自带，也可以用y_placeholder设置默认值
            单样本：y, y_hat
            多样本：[(y1, y_hat1), (y2, y_hat2), ...]
        :return:
            单样本：y_hat
            多样表：[y_hat1, y_hat2, ...]

        根据不同model结构特殊性
        """
        loader = self.inputs2loader(raw_in, batch_size=batch_size, y_placeholder=y_placeholder)
        preds = self.forward(loader, info=info, return_gt=return_gt)
        # 返回结果，单样本的时候作简化
        if len(preds) == 1 and not isinstance(raw_in, (list, tuple, set)):
            return preds[0]
        else:
            return preds


def setup_seed(seed):
    """ 完整的需要设置的随机数种子

    不过个人实验有时候也不一定有用~~
    还是有可能各种干扰因素导致模型无法复现
    """
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)
    random.seed(seed)
    torch.backends.cudnn.benchmark = False
    torch.backends.cudnn.deterministic = True


class TrainingSampler:
    """ 摘自detectron2，用来做无限循环的抽样
    我这里的功能做了简化，只能支持单卡训练，原版可以支持多卡训练

    In training, we only care about the "infinite stream" of training data.
    So this sampler produces an infinite stream of indices and
    all workers cooperate to correctly shuffle the indices and sample different indices.

    The samplers in each worker effectively produces `indices[worker_id::num_workers]`
    where `indices` is an infinite stream of indices consisting of
    `shuffle(range(size)) + shuffle(range(size)) + ...` (if shuffle is True)
    or `range(size) + range(size) + ...` (if shuffle is False)
    """

    def __init__(self, size: int, shuffle: bool = True):
        """
        Args:
            size (int): the total number of data of the underlying dataset to sample from
            shuffle (bool): whether to shuffle the indices or not
        """
        self._size = size
        assert size > 0
        self._shuffle = shuffle

    def __iter__(self):
        g = torch.Generator()
        while True:
            if self._shuffle:
                yield from torch.randperm(self._size, generator=g).tolist()
            else:
                yield from torch.arange(self._size).tolist()


class ZcPredictor:
    """ 智财ocrwork框架的封装接口

    这个本来是特用功能，不应该放这里的，反正也没啥不可公开的技术细节，为了使用方便就放这了
    """

    def __init__(self, config_file, *, gpu=None, batch_size=None, opts=None):
        """
        :param config_file: 支持输入配置文件路径，或者字符串格式的配置参数值
        :param gpu: 默认可以不设，会挑选当前最大剩余的一张卡
            注意配置文件中也有gpu参数，在该接口模式下会被弃用
        :param batch_size: 每次能同时识别的最大图片数
            注意config_file里也有batch_size，不过那是训练用的参数，跟这没必然联系，部署最好额外设置batch_size
            该参数可以不设，默认每次传入多少张图，就同时多少张进行批处理
        :param opts: 除了配置文件的参数，可以自设字典，覆盖更新配置参数值，常用的参数有
        """
        # 1 配置参数
        if isinstance(config_file, str) and config_file[-5:].lower() == '.yaml':
            deploy_path = os.environ.get('OCRWORK_DEPLOY', '.')  # 支持在环境变量自定义：部署所用的配置、模型所在目录
            config_file = os.path.join(deploy_path, config_file)
            f = open(config_file, "r")
        elif isinstance(config_file, str):
            f = io.StringIO(config_file)
        else:
            raise TypeError
        prepare_args = EasyDict(list(yaml.load_all(f, Loader=yaml.FullLoader))[0])
        f.close()

        # 2 特殊配置参数
        opts = opts or {}
        if gpu is not None:
            opts['gpu'] = str(gpu)
        if 'gpu' not in opts:
            # gpu没设置的时候，默认找一个空闲最大的显卡
            opts['gpu'] = NvmDevice().get_most_free_gpu_id()
        if 'gpu' in opts:  # 智财的配置好像必须要写字符串
            opts['gpu'] = str(opts['gpu'])
        prepare_args.update(opts)

        # 3 初始化各组件
        self.prepare_args = prepare_args
        self.batch_size = batch_size
        self.transform = lambda x: xlcv.read(x, 1)  # 默认统一转cv2的图片格式
        # self.transform = lambda x: PilPrcs.read(x, 1)  # 也可以使用pil图片格式

    def forward(self, imgs):
        raise NotImplemented('子类必须实现forward方法')

    def __call__(self, raw_in, *, batch_size=None, progress=False):
        """ 智财的框架，dataloader默认不需要对齐，重置collate_fn
        （其实不是不需要对齐，而是其augument组件会处理）

        :return: 以多个结果为例
            preds结果是list
            pred = preds[0]
                pred也是list，是第0张图的所有检测框，比如一共8个
                    每个框是 4*2 的numpy矩阵（整数）
        """
        # 1 判断长度
        if not getattr(raw_in, '__len__', None):
            imgs = [raw_in]
        else:
            imgs = raw_in
        n = len(imgs)
        batch_size = first_nonnone([batch_size, self.batch_size, n])

        # 2 一段一段处理
        preds = []
        t = tqdm(desc='forward', total=n, disable=not progress)
        for i in range(0, n, batch_size):
            inputs = imgs[i:i + batch_size]
            preds += self.forward([self.transform(img) for img in inputs])
            t.update(len(inputs))

        # 3 返回结果，单样本的时候作简化
        if len(preds) == 1 and not getattr(raw_in, '__len__', None):
            return preds[0]
        else:
            return preds
